'''
EM1d : 1D Frequency domain Maxwell equation.

    This module is meant to solve 

    (curl v, curl u) + (v, e u) 
           - (v, n x (1/mu curl u)_s = i w (jext, u) 

    in 1D. x is the direction of the wave propagation.
    Ex, Ey, Ez are treated as L2, H1, H1 elements respectively.

  Domain:   
     EM1D_Anisotropic : tensor dielectric
     EM1D_Vac         : scalar dielectric

  Boundary:
     EM1D_PEC         : Perfect electric conductor
     EM1D_PMC         : Perfect magnetic conductor
     EM1D_H           : Mangetic field boundary
     EM1D_Port        : Surface current
     EM1D_E           : Electric field
     EM1D_Continuity  : Continuitiy

  Note: the above modules are plan. Not fully implemented.

'''
from petram.phys.vtable import VtableElement, Vtable
import numpy as np
import traceback

from petram.model import Domain, Bdry, Pair
from petram.phys.phys_model import Phys
from petram.phys.common.em_base import EMPhysModule
from petram.phys.em1d.em1d_base import EM1D_Bdry
from petram.phys.em1d.em1d_vac import EM1D_Vac

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM1DModel')


class EM1D_DefDomain(EM1D_Vac):
    can_delete = False
    nlterms = []
    # vt  = Vtable(data1)
    # do not use vtable here, since we want to use
    # vtable defined in EM3D_Vac in add_bf_conttribution

    def __init__(self, **kwargs):
        super(EM1D_DefDomain, self).__init__(**kwargs)

    def panel1_param(self):
        return [['Default Domain (Vac)',   "eps_r=1, mu_r=1, sigma=0, ky=0, kz=0",  2, {}], ]

    def get_panel1_value(self):
        return ["eps_r=1, mu_r=1, sigma=0", ]

    def import_panel1_value(self, v):
        pass

    def panel1_tip(self):
        return None

    def get_possible_domain(self):
        return []


data2 = (('label1', VtableElement(None,
                                  guilabel='Default Bdry (PMC)',
                                  default="Ht = 0",
                                  tip="this is a natural BC")),)


class EM1D_DefBdry(EM1D_Bdry):
    can_delete = False
    is_essential = False
    nlterms = []
    vt = Vtable(data2)

    def __init__(self, **kwargs):
        super(EM1D_DefBdry, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(EM1D_DefBdry, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = ['remaining']
        return v

    def get_possible_bdry(self):
        return []


class EM1D_DefPair(Pair, Phys):
    can_delete = False
    is_essential = False
    is_complex = True

    def __init__(self, **kwargs):
        super(EM1D_DefPair, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(EM1D_DefPair, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    def get_possible_pair(self):
        return []


class EM1D(EMPhysModule):
    der_vars_base = ['B']
    geom_dim = 1

    def __init__(self, **kwargs):
        super(EM1D, self).__init__(**kwargs)
        self['Domain'] = EM1D_DefDomain()
        self['Boundary'] = EM1D_DefBdry()
        self['Pair'] = EM1D_DefPair()

    @property
    def dep_vars(self):
        '''
        list of dependent variables, for example.
           [Et, rEf]      
           [Et, rEf, psi]
        '''
        coord = ['x', 'y', 'z']
        ret = self.dep_vars_base + self.dep_vars_suffix
        return [ret + x for x in coord]

    @property
    def der_vars(self):
        coord = ['x', 'y', 'z']
        ret = EM1D.der_vars_base[0] + self.dep_vars_suffix
        return [ret + x for x in coord]

    @property
    def dep_vars0(self):
        coord = ['x', 'y', 'z']
        ret = self.dep_vars_base + self.dep_vars_suffix
        return [ret + x for x in coord]

    @property
    def dep_vars_base(self):
        ret = 'E'
        return ret

    def get_fec_type(self, idx):
        '''
        H1 
        H1v2 (vector dim)
        ND
        RT
        '''
        if self.use_h1_x:
            values = ['H1', 'H1', 'H1']
        else:
            values = ['L2', 'H1', 'H1']
        return values[idx]

    def get_fec(self):
        v = self.dep_vars
        if self.use_h1_x:
            return [(v[0], 'H1_FECollection'),
                    (v[1], 'H1_FECollection'),
                    (v[2], 'H1_FECollection'), ]
        return [(v[0], 'L2_FECollection'),
                (v[1], 'H1_FECollection'),
                (v[2], 'H1_FECollection'), ]

    def fec_order(self, idx):
        self.vt_order.preprocess_params(self)
        if idx == 0:
            if self.use_h1_x:
                return self.order
            return self.order - 1
        else:
            return self.order

    def _has_div_constraint(self):
        return False
        # from .em1d_div import EM1D_Div
        # for mm in self['Domain'].iter_enabled():
        #    if isinstance(mm, EM1D_Div): return True
        # return False

    def attribute_set(self, v):
        v = super(EM1D, self).attribute_set(v)
        v["element"] = 'L2_FECollection, H1_FECollection, H1_FECollection'
        v["ndim"] = 1
        v["ind_vars"] = 'x'
        v["dep_vars_suffix"] = ''
        v["use_h1_x"] = False
        return v

    def panel1_param(self):
        panels = super(EM1D, self).panel1_param()
        panels.extend([  # self.make_param_panel('freq',  self.freq_txt),
            ["independent vars.", self.ind_vars, 0, {}],
            ["dep. vars. suffix", self.dep_vars_suffix, 0, {}],
            ["dep. vars.", ','.join(self.dep_vars), 2, {}],
            ["ns vars.", ','.join(self.der_vars), 2, {}],
            ["use H1 for Ex", self.use_h1_x, 3, {"text": ' '}], ])
        return panels

    def get_panel1_value(self):
        names = ','.join([x for x in self.dep_vars])
        names2 = ', '.join(list(self.get_default_ns()))
        val = super(EM1D, self).get_panel1_value()
        # val.extend([self.freq_txt, self.ind_vars, self.dep_vars_suffix,
        #            names, names2, self.use_h1_x])
        val.extend([self.ind_vars, self.dep_vars_suffix,
                    names, names2, self.use_h1_x])
        return val

    def import_panel1_value(self, v):
        v = super(EM1D, self).import_panel1_value(v)

        self.ind_vars = str(v[0])
        self.dep_vars_suffix = str(v[1])
        self.use_h1_x = v[-1]

        if self.use_h1_x:
            self.element = 'H1_FECollection, H1_FECollection, H1_FECollection'
        else:
            self.element = 'L2_FECollection, H1_FECollection, H1_FECollection'

    def get_possible_domain(self):
        if EM1D._possible_constraints is None:
            self._set_possible_constraints('em1d')

        doms = super(EM1D, self).get_possible_domain()
        return EM1D._possible_constraints['domain'] + doms

    def get_possible_bdry(self):
        if EM1D._possible_constraints is None:
            self._set_possible_constraints('em1d')
        bdrs = super(EM1D, self).get_possible_bdry()
        return EM1D._possible_constraints['bdry'] + bdrs

    def get_possible_edge(self):
        return []

    def get_possible_pair(self):
        # from em3d_floquet       import EM3D_Floquet
        return []

    def get_possible_point(self):
        return []

    def add_variables(self, v, name, solr, soli=None):
        from petram.helper.variables import add_coordinates
        from petram.helper.variables import add_scalar
        from petram.helper.variables import add_components
        from petram.helper.variables import add_component_expression as addc_expression
        from petram.helper.variables import add_expression
        from petram.helper.variables import add_surf_normals
        from petram.helper.variables import add_constant

        from petram.phys.em1d.eval_deriv import eval_grad

        v = super(EM1D, self).add_variables(v, name, solr, soli)

        ind_vars = [x.strip() for x in self.ind_vars.split(',')]
        suffix = self.dep_vars_suffix

        from petram.helper.variables import TestVariable

        add_coordinates(v, ind_vars)
        add_surf_normals(v, ind_vars)
        add_scalar(v, name, "", ind_vars, solr, soli)

        if name.startswith('E'):
            if name.endswith('y'):
                add_scalar(v, 'gradEy', suffix, ind_vars, solr, soli,
                           deriv=eval_grad, vars=['E'])
            if name.endswith('z'):
                add_scalar(v, 'gradEz', suffix, ind_vars, solr, soli,
                           deriv=eval_grad, vars=['E'])

        addc_expression(v, 'B', suffix, ind_vars,
                        '-1j/omega*(1j*ky*Ez - 1j*kz*Ey)',
                        ['ky', 'kz', 'E', 'omega'], 0)
        addc_expression(v, 'B', suffix, ind_vars,
                        '-1j/omega*(1j*kz*Ex - gradEz)',
                        ['ky', 'kz', 'E', 'omega'], 'y')
        addc_expression(v, 'B', suffix, ind_vars,
                        '-1j/omega*(gradEy - 1j*ky*Ex)',
                        ['ky', 'kz', 'E', 'omega'], 'z')
        add_expression(v, 'E', suffix, ind_vars,
                       'array([Ex, Ey, Ez])',  ['E'])

        add_expression(v, 'B', suffix, ind_vars,
                       'array([Bx, By, Bz])',  ['B'])

        # Poynting Flux
        addc_expression(v, 'Poy', suffix, ind_vars,
                        '(conj(Ey)*Bz - conj(Ez)*By)/mu0',
                        ['B', 'E'], 0)
        addc_expression(v, 'Poy', suffix, ind_vars,
                        '(conj(Ez)*Bx - conj(Ex)*Bz)/mu0',
                        ['B', 'E'], 'y')
        addc_expression(v, 'Poy', suffix, ind_vars,
                        '(conj(Ex)*By - conj(Ey)*Bx)/mu0',
                        ['B', 'E'], 'z')

        return v

    def get_fes_for_dep(self, unknown_name, soldict):
        keys = soldict.keys()

        print(unknown_name, soldict.keys())
        assert False,  "check if this is called"
        k = unknown_name
        sol = soldict[k]
        solr = sol[0]
        soli = sol[1] if len(sol) > 1 else None
        return solr, soli



'''
  continuity is a condition for internail
  boundary. 
  It has no contribution to weak form
'''
from petram.model import Domain, Bdry, Pair
from petram.phys.phys_cont  import PhysContinuity

def bdry_constraints():
   return [EM3D_Continuity]

class EM3D_Continuity(PhysContinuity):
    pass




'''
EM2Da : Axis- symmetric Frequency domain Maxwell equation.

    This module is meant to solve

    (curl v, curl u) + (v, e u)
           - (v, n x (1/mu curl u)_s = i w (jext, u)

    We decompose the r-phi-z filed coponents in toroidal and poloidal
    components, Et and Ep. Note that Ep is a vector and Et is not Ephi

    Ep = (Er, Ez) and Et= rEphi, where Ep is Nedelec and Et is H1

    curl-curl is decomposed to
       r (curl Ep, curl Wp) + 1/r (m Et - div Ep,  m Wt - div Wp),

    mass
       e (r (Ep, Wp) + 1/r (Et, Wt)),

    where e = - epsion * w^2 - i * sigma * w, w is frequency,
    and  mu, epislon, and simga is regular EM parameters.

    ( , ) is volume integral and ( , )_s is surface intengral
    on non-essential boundaries.

    exp(-i w t) is assumed.

 *sigma
  Domain:
     EM2Da_Anisotropic : tensor dielectric
     EM2Da_Vac         : scalar dielectric
     EM2Da_ExtJ        : external current
     EM2Da_Div         : div J = 0 constraints (add Lagrange multiplier)

  Boundary:
     EM2Da_PEC         : Perfect electric conductor
     EM2Da_PMC         : Perfect magnetic conductor
     EM2Da_H           : Mangetic field boundary
     EM2Da_SurfJ       : Surface current
     EM2Da_Port        : TE, TEM, Coax port
     EM2Da_E           : Electric field
     EM2Da_Continuity  : Continuitiy

'''
from petram.phys.vtable import VtableElement, Vtable
import numpy as np
import traceback

from petram.model import Domain, Bdry, Pair
from petram.phys.phys_model import Phys
from petram.phys.common.em_base import EMPhysModule
from petram.phys.em2da.em2da_base import EM2Da_Bdry
from petram.phys.em2da.em2da_vac import EM2Da_Vac

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2DaModel')


class EM2Da_DefDomain(EM2Da_Vac):
    can_delete = False
    nlterms = []
    # vt  = Vtable(data1)
    # do not use vtable here, since we want to use
    # vtable defined in EM3D_Vac in add_bf_conttribution

    def __init__(self, **kwargs):
        super(EM2Da_DefDomain, self).__init__(**kwargs)

    def panel1_param(self):
        return [['Default Domain (Vac)',   "eps_r=1, mu_r=1, sigma=0",  2, {}], ]

    def get_panel1_value(self):
        return ["eps_r=1, mu_r=1, sigma=0", ]

    def import_panel1_value(self, v):
        pass

    def panel1_tip(self):
        return None

    def get_possible_domain(self):
        return []


data2 = (('label1', VtableElement(None,
                                  guilabel='Default Bdry (PMC)',
                                  default="Ht = 0",
                                  tip="this is a natural BC")),)


class EM2Da_DefBdry(EM2Da_Bdry):
    can_delete = False
    is_essential = False
    nlterms = []
    vt = Vtable(data2)

    def __init__(self, **kwargs):
        super(EM2Da_DefBdry, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(EM2Da_DefBdry, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = ['remaining']
        return v

    def get_possible_bdry(self):
        return []


class EM2Da_DefPair(Pair, Phys):
    can_delete = False
    is_essential = False
    is_complex = True

    def __init__(self, **kwargs):
        super(EM2Da_DefPair, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(EM2Da_DefPair, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    def get_possible_pair(self):
        return []


class EM2Da(EMPhysModule):
    der_vars_base = ['Er', 'Ephi', 'Ez', 'Br', 'Bphi', 'Bz', 'm_mode']
    der_vars_vec = ['E', 'B']
    geom_dim = 2

    def __init__(self, **kwargs):
        super(EM2Da, self).__init__(**kwargs)
        self['Domain'] = EM2Da_DefDomain()
        self['Boundary'] = EM2Da_DefBdry()
        self['Pair'] = EM2Da_DefPair()

    @property
    def dep_vars(self):
        '''
        list of dependent variables, for example.
           [Et, rEf]
           [Et, rEf, psi]
        '''
        ret = self.dep_vars_base
        return [x + self.dep_vars_suffix for x in ret]

    @property
    def dep_vars0(self):
        '''
        list of dependent variables, for example.
           [Et, rEf]
           [Et, rEf, psi]
        '''
        ret = self.dep_vars_base
        return [x + self.dep_vars_suffix for x in ret]

    @property
    def dep_vars_base(self):
        if self._has_div_constraint():
            ret = ['Et', 'rEf', 'psi']
        else:
            ret = ['Et', 'rEf']
        return ret

    def get_fec_type(self, idx):
        values = ['ND', 'H1', 'H1']
        return values[idx]

    def get_fec(self):
        v = self.dep_vars
        if len(v) == 2:  # normal case
            return [(v[0], 'ND_FECollection'),
                    (v[1], 'H1_FECollection'), ]
        else:            # with divergence constraints
            return [(v[0], 'ND_FECollection'),
                    (v[1], 'H1_FECollection'),
                    (v[2], 'H1_FECollection'), ]

    def _has_div_constraint(self):
        return False
        # from .em2da_div import EM2Da_Div
        # for mm in self['Domain'].iter_enabled():
        #    if isinstance(mm, EM2Da_Div): return True
        # return False

    def attribute_set(self, v):
        v = super(EM2Da, self).attribute_set(v)
        v["element"] = 'ND_FECollection, H1_FECollection'
        v["ndim"] = 2
        v["ind_vars"] = 'r, z'
        v["dep_vars_suffix"] = ''
        return v

    def panel1_param(self):
        panels = super(EM2Da, self).panel1_param()
        panels.extend([["independent vars.", self.ind_vars, 0, {}],
                       ["dep. vars. suffix", self.dep_vars_suffix, 0, {}],
                       ["dep. vars.", ','.join(self.dep_vars), 2, {}],
                       ["ns vars.", ','.join(EM2Da.der_vars_base), 2, {}], ])
        return panels

    def get_panel1_value(self):
        names = ', '.join([x for x in self.dep_vars])
        names2 = ', '.join(list(self.get_default_ns()))
        val = super(EM2Da, self).get_panel1_value()
        # val.extend([self.freq_txt, self.ind_vars, self.dep_vars_suffix,
        #            names, names2, ])
        val.extend([self.ind_vars, self.dep_vars_suffix,
                    names, names2, ])

        return val

    def import_panel1_value(self, v):
        v = super(EM2Da, self).import_panel1_value(v)
        # self.freq_txt = str(v[0])
        self.ind_vars = str(v[0])
        self.dep_vars_suffix = str(v[1])

    def get_possible_domain(self):
        if EM2Da._possible_constraints is None:
            self._set_possible_constraints('em2da')

        doms = super(EM2Da, self).get_possible_domain()
        return EM2Da._possible_constraints['domain'] + doms

    def get_possible_bdry(self):
        if EM2Da._possible_constraints is None:
            self._set_possible_constraints('em2da')
        bdrs = super(EM2Da, self).get_possible_bdry()
        return EM2Da._possible_constraints['bdry'] + bdrs

    def get_possible_edge(self):
        return []

    def get_possible_pair(self):
        if EM2Da._possible_constraints is None:
            self._set_possible_constraints('em2da')

        pairs = super(EM2Da, self).get_possible_pair()
        return EM2Da._possible_constraints['pair'] + pairs

    def get_possible_point(self):
        return []

    def add_variables(self, v, name, solr, soli=None):
        from petram.helper.variables import add_coordinates
        from petram.helper.variables import add_scalar
        from petram.helper.variables import add_components
        from petram.helper.variables import add_elements
        from petram.helper.variables import add_expression
        from petram.helper.variables import add_component_expression as addc_expression
        from petram.helper.variables import add_surf_normals
        from petram.helper.variables import add_constant

        from petram.phys.em2da.eval_deriv import eval_curl, eval_grad

        v = super(EM2Da, self).add_variables(v, name, solr, soli)

        def eval_curlEt(gfr, gfi=None):
            gfr, gfi, extra = eval_curl(gfr, gfi)
            return gfr, gfi, extra

        def eval_gradrEf(gfr, gfi=None):
            gfr, gfi, extra = eval_grad(gfr, gfi)
            return gfr, gfi, extra

        ind_vars = [x.strip() for x in self.ind_vars.split(',')]
        suffix = self.dep_vars_suffix

        add_coordinates(v, ind_vars)
        add_surf_normals(v, ind_vars)

        if name.startswith('Et'):
            add_elements(v, 'E', suffix, ind_vars, solr, soli, elements=[0, 1])
            add_scalar(v, 'curlEt', suffix, ind_vars, solr, soli,
                       deriv=eval_curlEt)
            addc_expression(v, 'B', suffix, ind_vars,
                            '-1j/omega*(-1)*curlEt', ['curlEt', 'omega'], 'phi')

        elif name.startswith('rEf'):
            add_scalar(v, 'rEf', suffix, ind_vars, solr, soli)
            addc_expression(v, 'E', suffix, ind_vars,
                            'rEf/r', ['rEf', ], 'phi')
            add_components(v, 'gradrE', suffix, ind_vars, solr, soli,
                           deriv=eval_gradrEf)
        elif name.startswith('psi'):
            add_scalar(v, 'psi', suffix, ind_vars, solr, soli)

        add_expression(v, 'E', suffix, ind_vars,
                       'array([Er, Ephi, Ez])',
                       ['E'])

        addc_expression(v, 'E', suffix, ind_vars,
                        'rEf/r', ['rEf', ], 'phi')
        addc_expression(v, 'B', suffix, ind_vars,
                        '-1j/omega*(1j*m_mode*Ez/r-gradrEz/r)',
                        ['m_mode', 'E', 'omega'], 0)
        addc_expression(v, 'B', suffix, ind_vars,
                        '-1j/omega*(-1j*m_mode*Er/r+gradrEr/r)',
                        ['m_mode', 'E', 'omega'], 1)
        add_expression(v, 'B', suffix, ind_vars,
                       'array([Br, Bphi, Bz])',
                       ['B'])

        # Poynting Flux
        addc_expression(v, 'Poy', suffix, ind_vars,
                        '(conj(Ephi)*Bz - conj(Ez)*Bphi)/mu0',
                        ['B', 'E'], 0)
        addc_expression(v, 'Poy', suffix, ind_vars,
                        '(conj(Ez)*Br - conj(Er)*Bz)/mu0',
                        ['B', 'E'], 'phi')
        addc_expression(v, 'Poy', suffix, ind_vars,
                        '(conj(Er)*Bphi - conj(Ephi)*Br)/mu0',
                        ['B', 'E'], 1)

        # collect all definition from children
        '''
        for mm in self.walk():
            if not mm.enabled: continue
            if mm is self: continue
            mm.add_domain_variables(v, name, suffix, ind_vars,
                                    solr, soli)
            mm.add_bdr_variables(v, name, suffix, ind_vars,
                                    solr, soli)
        '''
        return v

    def get_fes_for_dep(self, unknown_name, soldict):
        keys = soldict.keys()
        for k in keys:
            if unknown_name.startswith('phi'):
                if k.startswith('psi'):
                    break
            elif unknown_name.startswith('Et'):
                if k.startswith('Et'):
                    break
            else:
                if k.startswith('rEf'):
                    break
        sol = soldict[k]
        solr = sol[0]
        soli = sol[1] if len(sol) > 1 else None
        return solr, soli



'''
   Vacuum region:
      However, can have arbitrary scalar epsilon_r, mu_r, sigma


'''
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.phys_const import mu0, epsilon0

from petram.mfem_config import use_parallel
import numpy as np


from petram.phys.phys_model  import PhysCoefficient, PhysConstant
from petram.phys.coefficient import SCoeff, MCoeff
from petram.phys.em1d.em1d_base import EM1D_Bdry, EM1D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM1D_Vac')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('epsilonr', VtableElement('epsilonr', type='complex',
                                   guilabel='epsilonr',
                                   default=1.0,
                                   tip="relative permittivity")),
        ('mur', VtableElement('mur', type='complex',
                              guilabel='mur',
                              default=1.0,
                              tip="relative permeability")),
        ('sigma', VtableElement('sigma', type='complex',
                                guilabel='sigma',
                                default=0.0,
                                tip="contuctivity")),
        ('ky', VtableElement('ky', type='float',
                             guilabel='ky',
                             default=0.,
                             no_func=True,
                             tip="wave number` in the y direction")),
        ('kz', VtableElement('kz', type='float',
                             guilabel='kz',
                             default=0.0,
                             no_func=True,
                             tip="wave number in the z direction")),)


'''       
class Epsilon(PhysCoefficient):
   #
   #  - omega^2 * epsilon0 * epsilonr
   #
   def __init__(self, *args, **kwargs):
       self.omega = kwargs.pop('omega', 1.0)
       super(Epsilon, self).__init__(*args, **kwargs)

class Sigma(PhysCoefficient):
   #
   # -1j * omega * sigma
   #
   def __init__(self, *args, **kwargs):
       self.omega = kwargs.pop('omega', 1.0)
       super(Sigma, self).__init__(*args, **kwargs)

   def EvalValue(self, x):
       v = super(Sigma, self).EvalValue(x)
       v = -1j * self.omega * v
       if self.real:  return v.real
       else: return v.imag

class InvMu(PhysCoefficient):
   #
   #   1/mu0/mur*(factor)
   #
   def __init__(self, *args, **kwargs):
       self._extra_fac = kwargs.pop("factor", 1.)
       super(InvMu, self).__init__(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        self._extra_fac = kwargs.pop("factor", 1.)
        super(InvMu, self).__init__(*args, **kwargs)

    def EvalValue(self, x):
        v = super(InvMu, self).EvalValue(x)
        v = 1/mu0/v*self._extra_fac

        if self.real:
            return v.real
        else:
            return v.imag

       if self.real:  return v.real
       else: return v.imag
'''
def Epsilon(exprs, ind_vars, l, g, omega=1, cnorm=1):
    # - omega^2 * epsilon0 * epsilonr
    fac = -epsilon0 * omega * omega * cnorm
    coeff = SCoeff([exprs], ind_vars, l, g, return_complex=True, scale=fac)
    return coeff

def Sigma(exprs, ind_vars, l, g, omega=1, cnorm=1):
    # - 1j * self.omega * sigma
    fac = - 1j * omega * cnorm
    coeff = SCoeff([exprs], ind_vars, l, g, return_complex=True, scale=fac)
    return coeff
       
def InvMu(m, ind_vars, l, g, factor=1, cnorm=1):
    coeff = SCoeff([m], ind_vars, l, g, return_complex=True, scale=mu0)
    return 1./coeff*factor*cnorm
       
def domain_constraints():
    return [EM1D_Vac]

class EM1D_Vac(EM1D_Domain):
    vt = Vtable(data)
    #nlterms = ['epsilonr']

    def has_bf_contribution(self, kfes):
        if kfes == 0:
            return True
        elif kfes == 1:
            return True
        elif kfes == 2:
            return True
        else:
            return False

    def has_mixed_contribution(self):
        return True

    def get_mixedbf_loc(self):
        '''
        r, c, and flag1, flag2 of MixedBilinearForm
           flag1 : take transpose if negative
           flag2 : take conj if -1
        '''
        return [(0, 1, 1, 1),
                (1, 0, 1, 1),
                (1, 2, 1, 1),
                (2, 1, 1, 1),
                (0, 2, 1, 1),
                (2, 0, 1, 1), ]

    def add_bf_contribution(self, engine, a, real=True, kfes=0,
                            ecsc=None):
        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        e, m, s, ky, kz = self.vt.make_value_or_expression(self)
        if not isinstance(e, str):
            e = str(e)
        if not isinstance(m, str):
            m = str(m)
        if not isinstance(s, str):
            s = str(s)

        txt = ["Ex in ", "Ey in ", "Ez in "][kfes]
        if real:
            dprint1("Add BF contribution(real) " + txt + str(self._sel_index))
        else:
            dprint1("Add BF contribution(imag)" + txt + str(self._sel_index))

        if ecsc is None:
            sc = Sigma(s, self.get_root_phys().ind_vars,
                       self._local_ns, self._global_ns,
                       omega=omega, cnorm=cnorm)
            ec = Epsilon(e, self.get_root_phys().ind_vars,
                                self._local_ns, self._global_ns,
                              omega=omega, cnorm=cnorm)
        else:
            # anistropic case...
            ec, sc = ecsc

        self.set_integrator_realimag_mode(real)
        
        self.add_integrator(engine, 'epsilonr', sc, a.AddDomainIntegrator,
                            mfem.MassIntegrator)
        self.add_integrator(engine, 'sigma', ec, a.AddDomainIntegrator,
                            mfem.MassIntegrator)
        if kfes == 0:  # Ex
            imu = InvMu(m,  self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            factor = ky**2 + kz**2, cnorm=cnorm)
            
            self.add_integrator(engine, 'mur', imu, a.AddDomainIntegrator,
                                mfem.MassIntegrator)

        elif kfes == 1 or kfes == 2: # Ey and Ez
            imu = InvMu(m,  self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns, cnorm=cnorm)

            self.add_integrator(engine, 'mur', imu, a.AddDomainIntegrator,
                                mfem.DiffusionIntegrator)

            if kfes == 1: fac = kz*kz
            if kfes == 2: fac = ky*ky            

            imu = InvMu(m,  self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            factor = fac, cnorm=cnorm)

            self.add_integrator(engine, 'mur', imu, a.AddDomainIntegrator,
                                mfem.MassIntegrator)

    def add_mix_contribution(self, engine, mbf, r, c, is_trans, real=True):

        if real:
            dprint1("Add mixed contribution(real)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))
        else:
            dprint1("Add mixed contribution(imag)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))

        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        e, m, s, ky, kz = self.vt.make_value_or_expression(self)

        if not isinstance(e, str):
            e = str(e)
        if not isinstance(m, str):
            m = str(m)
        if not isinstance(s, str):
            s = str(s)

        if r == 0 and c == 1:
            imu = InvMu(m,  self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        factor=1j*ky, cnorm=cnorm)
            itg = mfem.MixedScalarDerivativeIntegrator
        elif r == 0 and c == 2:
            imu = InvMu(m,  self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        factor=1j*kz, cnorm=cnorm)
            itg = mfem.MixedScalarDerivativeIntegrator
        elif r == 1 and c == 0:
            imu = InvMu(m,  self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        factor=1j*ky, cnorm=cnorm)
            itg = mfem.MixedScalarWeakDerivativeIntegrator
        elif r == 1 and c == 2:
            imu = InvMu(m,  self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        factor=-ky*kz, cnorm=cnorm)
            itg = mfem.MixedScalarMassIntegrator
        elif r == 2 and c == 0:
            imu = InvMu(m,  self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        factor=1j*kz, cnorm=cnorm)
            itg = mfem.MixedScalarWeakDerivativeIntegrator
        elif r == 2 and c == 1:
            imu = InvMu(m,  self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        factor=-ky*kz, cnorm=cnorm)
            itg = mfem.MixedScalarMassIntegrator
        else:
            assert False, "Something is wrong..if it comes here;D"

        self.set_integrator_realimag_mode(real)            
        self.add_integrator(engine, 'mur', imu,
                            mbf.AddDomainIntegrator, itg)

    def add_domain_variables(self, v, n, suffix, ind_vars):
        from petram.helper.variables import add_constant

        e, m, s, ky, kz = self.vt.make_value_or_expression(self)

        if len(self._sel_index) == 0:
            return

        add_constant(v, 'ky', suffix, np.float64(ky),
                     domains=self._sel_index,
                     gdomain=self._global_ns)

        add_constant(v, 'kz', suffix, np.float64(kz),
                     domains=self._sel_index,
                     gdomain=self._global_ns)

        self.do_add_scalar_expr(v, suffix, ind_vars,
                                'sepsilonr', e, add_diag=3)
        self.do_add_scalar_expr(v, suffix, ind_vars, 'smur', m, add_diag=3)
        self.do_add_scalar_expr(v, suffix, ind_vars, 'ssigma', s, add_diag=3)

        var = ['x', 'y', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')



'''
   Surface Current Boundary Condition

   On external surface    n \times H = J_surf
   On internal surface    n \times (H1 - H2) = J_surf (not tested)

   (note)
    In MKSA,  1/mu curl E = -dB/dt 1/mu= i omega B / mu .
      n \times 1/mu curl E = n \times -dB/dt 1/mu
                           = n \times i omega H 
                           = i omega J_surf
    
 
    Therefore, surface integral 
        \int W \dot (n \times 1/mu curl E) d\Omega
    becomes
         \int W \dot (i omega J_surf) d\Omega


   CopyRight (c) 2016-  S. Shiraiwa
'''
from petram.phys.vtable import VtableElement, Vtable
from petram.mfem_config import use_parallel
import numpy as np

from petram.phys.phys_const import mu0, epsilon0

from petram.phys.phys_model import (VectorPhysCoefficient,
                                    PhysCoefficient)
from petram.phys.em2da.em2da_base import EM2Da_Bdry

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2Da_SurfJ')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('surfJ', VtableElement('surfJ', type='complex',
                                guilabel='Surface J',
                                suffix=('r', 'phi', 'z'),
                                default=[0, 0, 0],
                                tip="surface current")),)


class rJsurfp(VectorPhysCoefficient):
    def __init__(self, *args, **kwargs):
        omega = kwargs.pop('omega', 1.0)
        self.mur = kwargs.pop('mur', 1.0)
        self.fac = -1j*omega
        super(rJsurfp, self).__init__(*args, **kwargs)

    def EvalValue(self, x):
        v = super(rJsurfp, self).EvalValue(x)
        v = np.array((v[0], v[2]))
        v = self.fac * v * x[0]
        if self.real:
            return v.real
        else:
            return v.imag


class Jsurft(PhysCoefficient):
    def __init__(self, *args, **kwargs):
        omega = kwargs.pop('omega', 1.0)
        self.mur = kwargs.pop('mur', 1.0)
        self.fac = -1j*omega
        super(rJsurfp, self).__init__(*args, **kwargs)

    def EvalValue(self, x):
        v = super(Jsurft, self).EvalValue(x)
        v = self.fac * v[1]
        if self.real:
            return v.real
        else:
            return v.imag


def bdry_constraints():
    return [EM2Da_SurfJ]


class EM2Da_SurfJ(EM2Da_Bdry):
    is_essential = False
    vt = Vtable(data)

    def has_lf_contribution(self, kfes=0):
        if kfes != 0:
            return False
        return True

    def add_lf_contribution(self, engine, b, real=True, kfes=0):
        if kfes != 0:
            return
        if real:
            dprint1("Add LF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add LF contribution(imag)" + str(self._sel_index))

        freq, omega = self.get_root_phys().get_freq_omega()
        f_name = self.vt.make_value_or_expression(self)

        if kfes == 0:
            coeff1 = rJsurfp(2, f_name[0],  self.get_root_phys().ind_vars,
                             self._local_ns, self._global_ns,
                             real=real, omega=omega,)
            self.add_integrator(engine, 'surfJ', coeff1,
                                b.AddBoundaryIntegrator,
                                mfem.VectorFEDomainLFIntegrator)

        elif kfes == 1:
            coeff1 = Jsurft(f_name[0],  self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            real=real, omega=omega,)
            self.add_integrator(engine, 'surfJ', coeff1,
                                b.AddBoundaryIntegrator,
                                mfem.DomainLFIntegrator)



'''
  Definition of 3D port modes

  TE Mode
    Expression are based on Microwave Engineering p122 - p123.
    Note that it consists from two terms
       1)  \int dS W \dot n \times iwH (VectorFETangentIntegrator does this)
       2)  an weighting to evaulate mode amplutude from the E field
           on a boundary
  TEM Mode
       E is parallel to the periodic edge
       Mode number is ignored (of course)

  About sign of port phasing.
       positive phasing means the incoming wave appears later (phase delay)
       at the port

  Note: jcH returns H/mu0 (which is B)
'''
from petram.mfem_config import use_parallel
from numpy import sin, cos, exp, sqrt, array, sum, pi, log, cross, arctan2

from petram.phys.phys_const import epsilon0, mu0

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_PortMode')

if use_parallel:
    import mfem.par as mfem
    from mfem.common.mpi_debug import nicePrint
    '''
    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid     = MPI.COMM_WORLD.rank
    '''
else:
    import mfem.ser as mfem
    nicePrint = dprint1

'''
   rectangular wg TE
'''


def get_portmode_coeff_cls(mode):
    if mode == 'TEM':
        return C_Et_TEM, C_jwHt_TEM
    elif mode == 'TE':
        return C_Et_TE, C_jwHt_TE
    elif mode == 'Coax(TEM)':
        return C_Et_CoaxTEM, C_jwHt_CoaxTEM
    elif mode == 'Circular(TE)':
        return C_Et_CircularTE, C_jwHt_CircularTE
    else:
        raise NotImplementedError(
            "you must implement this mode")


def TE_norm(m, n, a, b, alpha, gamma):
    if m == 0:
        return sqrt(a*b * alpha*gamma/8*n*n/b/b*2)
    elif n == 0:
        return sqrt(a*b * alpha*gamma/8*m*m/a/a*2)
    else:
        return sqrt(a*b * alpha*gamma/8*(m*m/a/a + n*n/b/b))


class C_Et_TE(mfem.VectorPyCoefficient):
    def __init__(self, sdim, bdry, real=True, eps=1.0, mur=1.0, amp=1.0, phase=0.0,
                 m=0, n=1):
        self.real = real
        self.a, self.b, self.c = bdry.a, bdry.b, bdry.c
        self.a_vec, self.b_vec = bdry.a_vec, bdry.b_vec
        self.m, self.n = m, n
        freq, omega = bdry.get_root_phys().get_freq_omega()

        k = eps*omega*sqrt(eps*epsilon0 * mur*mu0)
        kc = sqrt((self.m*pi/bdry.a)**2 +
                  (self.n*pi/bdry.b)**2)
        beta = sqrt(k**2 - kc**2)
        # self.AA = 1.0 ##
        alpha = omega*mur*mu0*pi/kc/kc
        gamma = beta*pi/kc/kc
        norm = TE_norm(self.m, self.n, self.a, self.b, alpha, gamma)
        self.AA = alpha/norm*amp * exp(1j*phase*pi/180.)
        dprint2("normalization to old ", self.AA)
        mfem.VectorPyCoefficient.__init__(self, sdim)

    def EvalValue(self, x):
        p = array(x)
        # x, y is positive (first quadrant)
        x = abs(sum((p - self.c)*self.a_vec))
        y = abs(sum((p - self.c)*self.b_vec))

        Ex = self.AA * self.n/self.b*(
            cos(self.m*pi*x/self.a) *
            sin(self.n*pi*y/self.b))
        Ey = - self.AA * self.m/self.a*(
            sin(self.m*pi*x/self.a) *
            cos(self.n*pi*y/self.b))

        E = Ex*self.a_vec + Ey*self.b_vec
        if self.real:
            return -E.real
        else:
            return -E.imag


class C_jwHt_TE(mfem.VectorPyCoefficient):
    def __init__(self, sdim, bdry, real=True, eps=1.0, mur=1.0, amp=1.0, phase=0.0,
                 cnorm=1.0, m=0, n=1):
        self.real = real
        self.phase = phase  # phase !=0 for incoming wave

        freq, omega = bdry.get_root_phys().get_freq_omega()
        # cnorm = bdry.get_root_phys().get_coeff_norm()

        self.a, self.b, self.c = bdry.a, bdry.b, bdry.c
        self.a_vec, self.b_vec = bdry.a_vec, bdry.b_vec
        self.m, self.n = m, n

        k = eps*omega*sqrt(eps*epsilon0 * mur*mu0)
        kc = sqrt((self.m*pi/bdry.a)**2 +
                  (self.n*pi/bdry.b)**2)
        if kc > k:
            raise ValueError('Mode does not propagate')
        beta = sqrt(k**2 - kc**2)
        dprint1("propagation constant:" + str(beta))

        alpha = omega*mur*mu0*pi/kc/kc
        gamma = beta*pi/kc/kc
        norm = TE_norm(self.m, self.n, self.a, self.b, alpha, gamma)
        self.AA = omega*gamma/norm*amp*cnorm

        # AA = omega*mur*mu0*pi/kc/kc*amp
        # self.AA = omega*beta*pi/kc/kc/AA
        # self.AA = omega*beta*pi/kc/kc*amp/1000.

        mfem.VectorPyCoefficient.__init__(self, sdim)

    def EvalValue(self, x):
        p = array(x)
        # x, y is positive (first quadrant)
        x = abs(sum((p - self.c)*self.a_vec))
        y = abs(sum((p - self.c)*self.b_vec))

        Hx = 1j*self.AA * self.m/self.a*(
            sin(self.m*pi*x/self.a) *
            cos(self.n*pi*y/self.b))
        Hy = 1j*self.AA * self.n/self.b*(
            cos(self.m*pi*x/self.a) *
            sin(self.n*pi*y/self.b))

        H = Hx*self.a_vec + Hy*self.b_vec

        H = H * exp(1j*self.phase/180.*pi)
        if self.real:
            return H.real
        else:
            return H.imag


'''
   rectangular port parallel metal TEM
'''


class C_Et_TEM(mfem.VectorPyCoefficient):
    def __init__(self, sdim, bdry, real=True, eps=1.0, mur=1.0, amp=1.0, phase=0.0, **kargs):
        mfem.VectorPyCoefficient.__init__(self, sdim)
        freq, omega = bdry.get_root_phys().get_freq_omega()
        self.real = real
        self.a_vec, self.b_vec = bdry.a_vec, bdry.b_vec
        self.AA = 1.0*amp * exp(1j*phase*pi/180.)

    def EvalValue(self, x):
        Ex = self.AA
        E = -Ex*self.b_vec
        if self.real:
            return -E.real
        else:
            return -E.imag


class C_jwHt_TEM(mfem.VectorPyCoefficient):
    def __init__(self, sdim, bdry, real=True, eps=1.0, mur=1.0, amp=1.0, phase=0.0, cnorm=1.0, **kargs):
        mfem.VectorPyCoefficient.__init__(self, sdim)
        freq, omega = bdry.get_root_phys().get_freq_omega()
        # cnorm = bdry.get_root_phys().get_coeff_norm()

        self.real = real
        self.phase = phase  # phase !=0 for incoming wave
        self.a_vec, self.b_vec = bdry.a_vec, bdry.b_vec
        self.AA = omega*sqrt(epsilon0*eps/mu0/mur)*amp*cnorm

    def EvalValue(self, x):
        Hy = 1j*self.AA
        H = Hy*self.a_vec
        H = H * exp(1j*self.phase*pi/180.)
        if self.real:
            return H.real
        else:
            return H.imag


'''
   coax port TEM
'''


def coax_norm(a, b, mur, eps):
    return sqrt(1/pi/sqrt(epsilon0*eps/mu0/mur)/log(b/a))


class C_Et_CoaxTEM(mfem.VectorPyCoefficient):
    def __init__(self, sdim, bdry, real=True, eps=1.0, mur=1.0, amp=1.0, phase=0.0, **kargs):
        mfem.VectorPyCoefficient.__init__(self, sdim)
        freq, omega = bdry.get_root_phys().get_freq_omega()

        self.real = real
        self.a = bdry.a
        self.b = bdry.b
        self.ctr = bdry.ctr
        self.AA = coax_norm(self.a, self.b, mur, eps) * \
            amp * exp(1j*phase*pi/180.)

    def EvalValue(self, x):
        r = (x - self.ctr)
        rho = sqrt(sum(r**2))
        nr = r/rho

#       E = nr*log(self.b/rho)/log(self.b/self.a)
#       E = nr/log(self.b/self.a)/rho
        E = nr/rho*self.AA
        if self.real:
            return E.real
        else:
            return E.imag


class C_jwHt_CoaxTEM(mfem.VectorPyCoefficient):
    def __init__(self, sdim, bdry, real=True, eps=1.0, mur=1.0, amp=1.0, phase=0.0, cnorm=1.0, **kargs):
        mfem.VectorPyCoefficient.__init__(self, sdim)
        freq, omega = bdry.get_root_phys().get_freq_omega()
        # cnorm = bdry.get_root_phys().get_coeff_norm()

        self.real = real
        self.norm = bdry.norm
        self.a = bdry.a
        self.b = bdry.b
        self.ctr = bdry.ctr
        self.phase = phase  # phase !=0 for incoming wave
        # self.AA = omega*sqrt(epsilon0*eps/mu0/mur)
        self.AA = coax_norm(self.a, self.b, mur, eps) * \
            omega*sqrt(epsilon0*eps/mu0/mur)*amp*cnorm

    def EvalValue(self, x):
        r = (x - self.ctr)
        rho = sqrt(sum(r**2))
        nr = r/rho
        nr = cross(nr, self.norm)
#       H = nr*log(self.b/rho)/log(self.b/self.a)
#       H = nr/log(self.b/self.a)/rho
        H = 1j*nr/rho*self.AA
        # H = 1j*self.AA*H
        H = H * exp(1j*self.phase*pi/180.)
        if self.real:
            return -H.real
        else:
            return -H.imag


'''
   circular port TEM
'''


def circular_norm():
    # ToDo
    return 1.0


class C_CircularTE(mfem.VectorPyCoefficient):
    def __init__(self, sdim, bdry, real=True, eps=1.0, mur=1.0, amp=1.0, phase=0.0, cnorm=1.0,
                 m=0, n=1):
        mfem.VectorPyCoefficient.__init__(self, sdim)
        freq, omega = bdry.get_root_phys().get_freq_omega()

        self.real = real
        self.m = m
        self.n = n
        self.a = bdry.a  #
        self.ctr = bdry.ctr
        self.ax1 = bdry.ax1
        self.ax2 = bdry.ax2
        self.norm = bdry.norm
        self.cnorm = cnorm
        self.phase = phase

        self.amp = amp

        freq, omega = bdry.get_root_phys().get_freq_omega()
        self.omega = omega

        from scipy.special import jnp_zeros

        xzero = jnp_zeros(abs(self.m), self.n)[-1]
        kc = xzero/self.a
        k = eps*omega*sqrt(epsilon0*eps * mu0*mur)
        if k**2 < kc**2:
            assert False, "mode does not propagte"

        self.kg = sqrt(k**2-kc**2)
        self.kc = kc
        #print("kc", self.kc)
        self.AA = circular_norm() * amp * exp(1j*phase*pi/180.)

    def compute_jv_x(self, rr):
        from scipy.special import jv
        from math import factorial

        m = abs(self.m)
        kc = self.kc

        if m > 0:
            x = rr*kc
            if x < 0.001:
                jv_x = 0
                for i in range(5):
                  jv_x += (-1)**i /factorial(i)/factorial(m+i) *x**(2*i + m-1)/2**(2*i + m)
            else:
                jv_x = jv(m, x)/x
            jv_x = jv_x*kc
        else:
            return 0

        if m < 0:
            jv_x*(-1)**m
        return jv_x

class C_Et_CircularTE(C_CircularTE):
    def EvalValue(self, x):
        r = (x - self.ctr)
        rr = sqrt(sum(r**2))
        th = arctan2(sum(r*self.ax2), sum(r*self.ax1))

        nr = self.ax1*cos(th) + self.ax2*sin(th)
        nt = -self.ax1*sin(th) + self.ax2*cos(th)

        jv_x = self.compute_jv_x(rr)

        from scipy.special import jvp
        
        if self.m > 0:
            Er = self.m * jv_x*exp(1j*self.m*th-1j*pi/2)
        else:
            Er = self.m * jv_x*exp(1j*self.m*th-1j*pi/2)*(-1)**self.m

        Et = self.kc*jvp(self.m, rr*self.kc)*exp(1j*self.m*th)


        E = (Er*nr + Et*nt) * self.AA

        if self.real:
            return E.real
        else:
            return E.imag


class C_jwHt_CircularTE(C_CircularTE):
    def EvalValue(self, x):
        r = (x - self.ctr)
        rr = sqrt(sum(r**2))
        th = arctan2(sum(r*self.ax2), sum(r*self.ax1))

        nr = self.ax1*cos(th) + self.ax2*sin(th)
        nt = -self.ax1*sin(th) + self.ax2*cos(th)

        jv_x = self.compute_jv_x(rr)

        from scipy.special import jvp

        # Hr = -self.kg*self.kc/self.omega*jvp(self.m, rr*self.kc)*cos(self.m*th)
        # Ht =  self.kg*self.m/self.omega/rr*jv(self.m, rr*self.kc)*sin(self.m*th)
        Hr = self.kg*self.kc/self.omega * \
            jvp(self.m, rr*self.kc)*exp(1j*self.m*th)
        #Ht = self.kg*self.m/self.omega/rr * \
        #    jv(self.m, rr*self.kc)*exp(1j*self.m*th)
        if self.m >= 0:
            Ht = -self.kg*self.m/self.omega*jv_x*exp(1j*self.m*th-1j*pi/2)
        else:
            Ht = -self.kg*self.m/self.omega*jv_x*exp(1j*self.m*th-1j*pi/2)*(-1)**self.m

        H = 1j*self.omega*(Hr*nr + Ht*nt) * self.AA*self.cnorm /mu0

        if self.real:
            return -H.real
        else:
            return -H.imag



'''
   cold plasma.
'''
from petram.phys.common.rf_dispersion_coldplasma import (stix_options,
                                                         default_stix_option,
                                                         default_stix_modelvalue,
                                                         col_model_options,
                                                         default_col_model,
                                                         vtable_data0)
import numpy as np

from petram.mfem_config import use_parallel, get_numba_debug

from petram.phys.vtable import VtableElement, Vtable
from petram.phys.phys_const import mu0, epsilon0
from petram.phys.numba_coefficient import NumbaCoefficient

from petram.phys.phys_model import MatrixPhysCoefficient, PhysCoefficient, PhysConstant, PhysMatrixConstant
from petram.phys.em1d.em1d_base import EM1D_Bdry, EM1D_Domain
from petram.phys.em1d.em1d_vac import EM1D_Vac

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM1D_ColdPlasma')

if use_parallel:
    import mfem.par as mfem
    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank
else:
    import mfem.ser as mfem
    myid = 0

vtable_data = vtable_data0.copy()
vtable_data.extend(
    [('ky', VtableElement('ky', type='float',
                          guilabel='ky',
                          default=0.,
                          no_func=True,
                          tip="wave number in the y direction")),
     ('kz', VtableElement('kz', type='float',
                          guilabel='kz',
                          default=0.0,
                          no_func=True,
                          tip="wave number in the z direction"))])


def domain_constraints():
    return [EM1D_ColdPlasma]


class EM1D_ColdPlasma(EM1D_Vac):
    allow_custom_intorder = False
    vt = Vtable(vtable_data)

    def __init__(self, **kargs):
        super(EM1D_ColdPlasma, self).__init__(**kargs)
        self._jited_coeff = None

    def get_possible_child(self):
        return []

    def attribute_set(self, v):
        EM1D_Vac.attribute_set(self, v)
        v["stix_terms"] = default_stix_modelvalue
        v["col_model"] = default_col_model
        return v

    def config_terms(self, evt):
        from petram.phys.common.rf_stix_terms_panel import ask_rf_stix_terms

        self.vt.preprocess_params(self)
        _B, _dens_e, _t_e, _dens_i, _masses, charges, _ky, _kz = self.vt.make_value_or_expression(
            self)

        num_ions = len(charges)
        win = evt.GetEventObject()
        value = ask_rf_stix_terms(win, num_ions, self.stix_terms)
        self.stix_terms = value

    def stix_terms_str(self):
        from petram.phys.common.rf_dispersion_coldplasma import value2modelstr
        return value2modelstr(self.stix_terms)

    def panel1_param(self):
        panels = super(EM1D_ColdPlasma, self).panel1_param()
        panels.append(["col. model", None, 1, {"values": col_model_options}])
        panels.extend([["Stix terms", "", 2, None],
                       [None, None, 341, {"label": "Customize terms",
                                          "func": "config_terms",
                                          "sendevent": True,
                                          "noexpand": True}], ])

        return panels

    def get_panel1_value(self):
        values = super(EM1D_ColdPlasma, self).get_panel1_value()
        values.extend([self.col_model, self.stix_terms_str(), self])
        return values

    def import_panel1_value(self, v):
        check = super(EM1D_ColdPlasma, self).import_panel1_value(v[:-3])
        self.col_model = v[-3]
        return check

    @property
    def jited_coeff(self):
        return self._jited_coeff

    def compile_coeffs(self):
        self._jited_coeff = self.get_coeffs()

    def get_coeffs(self):
        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        B, dens_e, t_e, dens_i, masses, charges, ky, kz = self.vt.make_value_or_expression(
            self)
        ind_vars = self.get_root_phys().ind_vars
        col_model = self.col_model

        from petram.phys.common.rf_dispersion_coldplasma import build_coefficients

        coeff1, coeff2, coeff3, coeff_nuei = build_coefficients(ind_vars, omega, B, dens_e, t_e,
                                                                        dens_i, masses, charges, col_model, cnorm,
                                                                        self._global_ns, self._local_ns,
                                                                        sdim=1, terms=self.stix_terms)
        return coeff1, coeff2, coeff3, coeff_nuei,  ky, kz

    def add_bf_contribution(self, engine, a, real=True, kfes=0):
        if real:
            dprint1("Add BF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add BF contribution(imag)" + str(self._sel_index))

        coeff1, coeff2, coeff3, _coeff_nuei,  ky, kz = self.jited_coeff
        self.set_integrator_realimag_mode(real)

        # if self.has_pml():
        #    coeff1 = self.make_PML_coeff(coeff1)
        #    coeff2 = self.make_PML_coeff(coeff2)
        #    coeff3 = self.make_PML_coeff(coeff3)

        ec = coeff1[kfes, kfes]
        sc = coeff3[kfes, kfes]

        self.add_integrator(engine, 'epsilonr', ec, a.AddDomainIntegrator,
                            mfem.MassIntegrator)
        self.add_integrator(engine, 'sigma', sc, a.AddDomainIntegrator,
                            mfem.MassIntegrator)

        coeff4 = 1./coeff2[0, 0]
        #
        #
        #
        if kfes == 0:  # Ex
            imu = coeff4*(ky**2 + kz**2)
            self.add_integrator(engine, 'mur', imu, a.AddDomainIntegrator,
                                mfem.MassIntegrator)

        elif kfes == 1 or kfes == 2:  # Ey and Ez
            imu = coeff4
            self.add_integrator(engine, 'mur1', imu, a.AddDomainIntegrator,
                                mfem.DiffusionIntegrator)

            if kfes == 1:
                fac = kz*kz
            if kfes == 2:
                fac = ky*ky

            imu = coeff4*fac

            self.add_integrator(engine, 'mur2', imu, a.AddDomainIntegrator,
                                mfem.MassIntegrator)

    def add_mix_contribution(self, engine, mbf, r, c, is_trans, real=True):
        if real:
            dprint1("Add mixed contribution(real)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))
        else:
            dprint1("Add mixed contribution(imag)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))

        coeff1, coeff2, coeff3, _coeff_nuei, ky, kz = self.jited_coeff
        self.set_integrator_realimag_mode(real)

        # super(EM1D_ColdPlasma, self).add_mix_contribution(engine, mbf, r, c, is_trans,
        #                                                   real = real)
        ec = coeff1[r, c]
        sc = coeff3[r, c]

        self.add_integrator(engine, 'epsilonr', ec, mbf.AddDomainIntegrator,
                            mfem.MixedScalarMassIntegrator)
        self.add_integrator(engine, 'sigma', sc, mbf.AddDomainIntegrator,
                            mfem.MixedScalarMassIntegrator)

        coeff4 = 1./coeff2[0, 0]
        if r == 0 and c == 1:
            imu = coeff4*(1j*ky)
            itg = mfem.MixedScalarDerivativeIntegrator
        elif r == 0 and c == 2:
            imu = coeff4*(1j*kz)
            itg = mfem.MixedScalarDerivativeIntegrator
        elif r == 1 and c == 0:
            imu = coeff4*(1j*ky)
            itg = mfem.MixedScalarWeakDerivativeIntegrator
        elif r == 1 and c == 2:
            imu = coeff4*(-ky*kz)
            itg = mfem.MixedScalarMassIntegrator
        elif r == 2 and c == 0:
            imu = coeff4*(1j*kz)
            itg = mfem.MixedScalarWeakDerivativeIntegrator
        elif r == 2 and c == 1:
            imu = coeff4*(-ky*kz)
            itg = mfem.MixedScalarMassIntegrator
        else:
            assert False, "Something is wrong..if it comes here;D"

        self.add_integrator(engine, 'mur', imu,
                            mbf.AddDomainIntegrator, itg)

    def add_domain_variables(self, v, n, suffix, ind_vars):
        from petram.helper.variables import add_constant

        if len(self._sel_index) == 0:
            return

        freq, omega = self.get_root_phys().get_freq_omega()
        B, dens_e, t_e, dens_i, masses, charges, ky, kz = self.vt.make_value_or_expression(
            self)
        col_model = self.col_model

        add_constant(v, 'ky', suffix, np.float64(kz),
                     domains=self._sel_index,
                     gdomain=self._global_ns)
        add_constant(v, 'kz', suffix, np.float64(kz),
                     domains=self._sel_index,
                     gdomain=self._global_ns)

        from petram.phys.common.rf_dispersion_coldplasma import build_variables

        ss = self.parent.parent.name()+'_'+self.name()  # phys module name + name
        ret  = build_variables(v, ss, ind_vars, omega, B, dens_e, t_e,
                               dens_i, masses, charges, col_model,
                               self._global_ns, self._local_ns,
                               terms=self.stix_terms,)

        from petram.phys.common.rf_dispersion_coldplasma import add_domain_variables_common
        add_domain_variables_common(self, ret, v, suffix, ind_vars)

        var = ['x', 'y', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')

        return



from petram.mfem_config import use_parallel
import numpy as np


from petram.phys.coefficient import VCoeff
from petram.phys.vtable import VtableElement, Vtable

#from petram.phys.phys_model  import Phys, VectorPhysCoefficient
from petram.phys.em3d.em3d_base import EM3D_Bdry

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_E')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('E', VtableElement('E', type='complex',
                            guilabel='Electric field',
                            suffix=('x', 'y', 'z'),
                            default=np.array([0, 0, 0]),
                            tip="essential BC")),)

def bdry_constraints():
    return [EM3D_E]


class EM3D_E(EM3D_Bdry):
    has_essential = True
    vt = Vtable(data)

    def attribute_set(self, v):
        super(EM3D_E, self).attribute_set(v)

        # old file may not contain E_x_txt
        if not hasattr(self, "E_x_txt") and hasattr(self, "E_x"):
            dprint1("making adjustment for backward compatilibty")
            setattr(self, "E_x_txt", str(self.E_x))
            setattr(self, "E_y_txt", str(self.E_y))
            setattr(self, "E_z_txt", str(self.E_z))
        return v

    def get_essential_idx(self, kfes):
        if kfes == 0:
            return self._sel_index
        else:
            return []

    def apply_essential(self, engine, gf, real=False, kfes=0):
        if kfes != 0:
            return
        if real:
            dprint1("Apply Ess.(real)" + str(self._sel_index))
        else:
            dprint1("Apply Ess.(imag)" + str(self._sel_index))

        Exyz = self.vt.make_value_or_expression(self)

        ind_vars = self.get_root_phys().ind_vars
        l = self._local_ns
        g = self._global_ns

        coeff1 = VCoeff(3, Exyz[0], ind_vars, l, g, return_complex=True)
        #f_name = self._make_f_name()
        # coeff1 = Et(3, f_name,  self.get_root_phys().ind_vars,
        #            self._local_ns, self._global_ns,
        #            real = real)
        coeff1 = self.restrict_coeff(coeff1, engine, vec=True)

        mesh = engine.get_mesh(mm=self)
        ibdr = mesh.bdr_attributes.ToList()
        bdr_attr = [0]*mesh.bdr_attributes.Max()
        for idx in self._sel_index:
            bdr_attr[idx-1] = 1

        coeff1 = coeff1.get_realimag_coefficient(real)
        gf.ProjectBdrCoefficientTangent(coeff1, mfem.intArray(bdr_attr))



'''
   local-K plasma.
'''
from petram.phys.common.rf_dispersion_lkplasma import (vtable_data0,
                                                       default_kpe_option,
                                                       kpe_options,
                                                       term_options,
                                                       default_term_option)

from petram.phys.common.rf_dispersion_coldplasma import (col_model_options,
                                                         default_col_model,)

import numpy as np

from petram.mfem_config import use_parallel, get_numba_debug

from petram.phys.vtable import VtableElement, Vtable
from petram.phys.phys_const import mu0, epsilon0
from petram.phys.coefficient import SCoeff, VCoeff, MCoeff
from petram.phys.numba_coefficient import NumbaCoefficient

from petram.phys.phys_model import MatrixPhysCoefficient, PhysCoefficient, PhysConstant, PhysMatrixConstant
from petram.phys.em3d.em3d_base import EM3D_Bdry, EM3D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_LocalKPlasma')

if use_parallel:
    import mfem.par as mfem
    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank
else:
    import mfem.ser as mfem
    myid = 0

vtable_data = vtable_data0.copy()

kpe_alg_options = ["std", ]


def domain_constraints():
    return [EM3D_LocalKPlasma]


class EM3D_LocalKPlasma(EM3D_Domain):
    allow_custom_intorder = True
    vt = Vtable(vtable_data)
    # nlterms = ['epsilonr']

    @classmethod
    def fancy_menu_name(cls):
        return "HotPlasma(local-k)"

    @classmethod
    def fancy_tree_name(cls):
        return "LocalKPlasma"

    def get_possible_child(self):
        from .em3d_pml import EM3D_LinearPML
        return [EM3D_LinearPML]

    def attribute_set(self, v):
        super(EM3D_LocalKPlasma, self).attribute_set(v)
        v["kpe_mode"] = default_kpe_option
        v["kpe_alg"] = kpe_alg_options[0]
        v["col_model"] = default_col_model
        v["lk_terms"] = None
        return v

    def config_terms(self, evt):
        from petram.phys.common.rf_lk_term_panel import ask_rf_lk_terms

        self.vt.preprocess_params(self)
        ret = self.vt.make_value_or_expression(self)
        charges = ret[7]

        num_ions = len(charges)
        win = evt.GetEventObject()
        value = ask_rf_lk_terms(win, num_ions, self.lk_terms)
        self.lk_terms = value

    def panel1_param(self):
        panels = super(EM3D_LocalKPlasma, self).panel1_param()
        panels.append(["kpe mode", None, 1, {"values": kpe_options}])
        panels.append(["kpe alg.", None, 1, {"values": kpe_alg_options}])
        panels.append(["col. model", None, 1, {"values": col_model_options}])
        panels.extend([  # ["contributions", "", 2, None],
            [None, None, 341, {"label": "Customize terms",
                               "func": "config_terms",
                               "sendevent": True,
                               "noexpand": True}], ])
        return panels

    def get_panel1_value(self):
        values = super(EM3D_LocalKPlasma, self).get_panel1_value()
        values.extend([self.kpe_mode, self.kpe_alg, self.col_model, self])
        return values

    def import_panel1_value(self, v):
        check = super(EM3D_LocalKPlasma, self).import_panel1_value(v[:-4])
        self.kpe_mode = v[-4]
        self.kpe_alg = v[-3]
        self.col_model = v[-2]
        return check

    @property
    def jited_coeff(self):
        return self._jited_coeff

    def compile_coeffs(self):
        self._jited_coeff = self.get_coeffs()

    def get_coeffs(self):
        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        B, dens_e, t_e, dens_i, t_i, t_c, masses, charges, kpakpe, kpevec = self.vt.make_value_or_expression(
            self)
        ind_vars = self.get_root_phys().ind_vars
        kpe_mode = self.kpe_mode
        kpe_alg = self.kpe_alg

        from petram.phys.common.rf_dispersion_lkplasma import value2flags

        num_ions = len(charges)
        terms = value2flags(num_ions, self.lk_terms)

        from petram.phys.common.rf_dispersion_lkplasma import build_coefficients

        coeff1, coeff2, coeff3, coeff4 = build_coefficients(ind_vars, omega, B, t_c, dens_e, t_e,
                                                            dens_i, t_i, masses, charges, kpakpe, kpevec,
                                                            kpe_mode, self.col_model, cnorm,
                                                            self._global_ns, self._local_ns,
                                                            kpe_alg=kpe_alg, sdim=3, terms=terms)

        return coeff1, coeff2, coeff3, coeff4

    def has_bf_contribution(self, kfes):
        if kfes == 0:
            return True
        else:
            return False

    def add_bf_contribution(self, engine, a, real=True, kfes=0):
        if kfes != 0:
            return
        if real:
            dprint1("Add BF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add BF contribution(imag)" + str(self._sel_index))

        coeff1, coeff2, coeff3, coeff4 = self.jited_coeff
        self.set_integrator_realimag_mode(real)

        if self.has_pml():
            coeff1 = self.make_PML_coeff(coeff1)
            coeff2 = self.make_PML_coeff(coeff2)
            coeff3 = self.make_PML_coeff(coeff3)
        coeff2 = coeff2.inv()

        if self.allow_custom_intorder and self.add_intorder != 0:
            fes = a.FESpace()
            geom = fes.GetFE(0).GetGeomType()
            order = fes.GetFE(0).GetOrder()
            isPK = (fes.GetFE(0).Space() == mfem.FunctionSpace.Pk)
            orderw = fes.GetElementTransformation(0).OrderW()
            curlcurl_order = order*2 - 2 if isPK else order*2
            mass_order = orderw + 2*order
            curlcurl_order += self.add_intorder
            mass_order += self.add_intorder

            dprint1("Debug: custom int order. Increment = " +
                    str(self.add_intorder))
            dprint1("  FE order: " + str(order))
            dprint1("  OrderW: " + str(orderw))
            dprint1("  CurlCurlOrder: " + str(curlcurl_order))
            dprint1("  FEMassOrder: " + str(mass_order))

            cc_ir = mfem.IntRules.Get(geom, curlcurl_order)
            ms_ir = mfem.IntRules.Get(geom, mass_order)
        else:
            cc_ir = None
            ms_ir = None

        self.add_integrator(engine, 'epsilonr', coeff1,
                            a.AddDomainIntegrator,
                            mfem.VectorFEMassIntegrator,
                            ir=ms_ir)

        if coeff2 is not None:
            self.add_integrator(engine, 'mur', coeff2,
                                a.AddDomainIntegrator,
                                mfem.CurlCurlIntegrator,
                                ir=cc_ir)
            # coeff2 = self.restrict_coeff(coeff2, engine)
            # a.AddDomainIntegrator(mfem.CurlCurlIntegrator(coeff2))
        else:
            dprint1("No contrinbution from curlcurl")

        self.add_integrator(engine, 'sigma', coeff3,
                            a.AddDomainIntegrator,
                            mfem.VectorFEMassIntegrator,
                            ir=ms_ir)

    def add_domain_variables(self, v, n, suffix, ind_vars):
        from petram.helper.variables import add_constant

        if len(self._sel_index) == 0:
            return

        freq, omega = self.get_root_phys().get_freq_omega()
        B, dens_e, t_e, dens_i, t_i, t_c, masses, charges, kpakpe, kpevec = self.vt.make_value_or_expression(
            self)

        kpe_mode = self.kpe_mode
        kpe_alg = self.kpe_alg

        from petram.phys.common.rf_dispersion_lkplasma import value2flags

        num_ions = len(charges)
        terms = value2flags(num_ions, self.lk_terms)

        from petram.phys.common.rf_dispersion_lkplasma import build_variables

        ss = self.parent.parent.name()+'_'+self.name()  # phys module name + name
        ret = build_variables(v, ss, ind_vars,
                              omega, B, t_c, dens_e, t_e,
                              dens_i, t_i, masses, charges,
                              kpakpe, kpevec, kpe_mode, kpe_alg, self.col_model,
                              self._global_ns, self._local_ns,
                              sdim=3, terms=terms)

        from petram.phys.common.rf_dispersion_lkplasma import add_domain_variables_common
        add_domain_variables_common(self, ret, v, suffix, ind_vars)

        var = ['x', 'y', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')

        return



'''
   Tangential H Boundary Condition

   (note)
    In MKSA,  1/mu curl E = -dH/dt = i omega H.
 
    Therefore, surface integral 
        \int W \dot (n \times 1/mu curl E) d\Omega
    becomes
        \int W \dot (n \times i omega H) d\Omega

    This terms goes to the right hand side. Thus. factor is - i omega

   CopyRight (c) 2016-  S. Shiraiwa
'''

from petram.phys.vtable import VtableElement, Vtable
from petram.mfem_config import use_parallel
import numpy as np

from petram.model import Domain, Bdry, Pair
from petram.phys.phys_model import Phys, VectorPhysCoefficient, PhysCoefficient
from petram.phys.em2d.em2d_base import EM2D_Bdry

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2D_H')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('H', VtableElement('H', type='complex',
                            guilabel='H',
                            suffix=('x', 'y', 'z'),
                            default=[0, 0, 0],
                            tip="magnetic field")),)


class Ht(VectorPhysCoefficient):
    def __init__(self, *args, **kwargs):
        omega = kwargs.pop('omega', 1.0)
        self.fac = -1j*omega  # /mur
        self.nor = None
        super(Ht, self).__init__(*args, **kwargs)

    def Eval(self, V, T, ip):
        nor = mfem.Vector(2)
        mfem.CalcOrtho(T.Jacobian(), nor)
        tmp = np.hstack([nor.GetDataArray(), 0.0])
        self.nor = tmp/np.linalg.norm(tmp)

        return VectorPhysCoefficient.Eval(self, V, T, ip)

    def EvalValue(self, x):
        v = super(Ht, self).EvalValue(x)
        nv = np.cross(self.nor, v)
        nv = self.fac * nv[:2]
        if self.real:
            return nv.real
        else:
            return nv.imag


class Hz(PhysCoefficient):
    def __init__(self, *args, **kwargs):
        omega = kwargs.pop('omega', 1.0)
        self.fac = -1j*omega  # /mur
        super(Hz, self).__init__(*args, **kwargs)

    def Eval(self, T, ip):
        nor = mfem.Vector(2)
        mfem.CalcOrtho(T.Jacobian(), nor)
        tmp = np.hstack([nor.GetDataArray(), 0.0])
        self.nor = tmp/np.linalg.norm(tmp)

        return PhysCoefficient.Eval(self, T, ip)

    def EvalValue(self, x):
        v = super(Hz, self).EvalValue(x)

        nv = np.cross(self.nor, v)
        nv = self.fac * nv[2]
        if self.real:
            return nv.real
        else:
            return nv.imag


def bdry_constraints():
    return [EM2D_H]


class EM2D_H(EM2D_Bdry):
    is_essential = False
    vt = Vtable(data)

    def has_lf_contribution(self, kfes=0):
        return True

    def add_lf_contribution(self, engine, b, real=True, kfes=0):
        if real:
            dprint1("Add LF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add LF contribution(imag)" + str(self._sel_index))

        from petram.phys.phys_const import mu0, epsilon0
        freq, omega = self.get_root_phys().get_freq_omega()

        h = self.vt.make_value_or_expression(self)
        if kfes == 0:
            coeff1 = Ht(2, h[0],  self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        real=real, omega=omega)
            self.add_integrator(engine, 'H', coeff1,
                                b.AddBoundaryIntegrator,
                                mfem.VectorFEDomainLFIntegrator)

        else:
            coeff1 = Hz(h[0],  self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        real=real, omega=omega)
            self.add_integrator(engine, 'H', coeff1,
                                b.AddBoundaryIntegrator,
                                mfem.BoundaryLFIntegrator)



'''`
   Essential BC
'''
import numpy as np

from petram.model import Bdry
from petram.phys.phys_model  import Phys, PhysCoefficient
from petram.phys.em1d.em1d_base import EM1D_Bdry, EM1D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM1D_E')

from petram.mfem_config import use_parallel
if use_parallel:
   import mfem.par as mfem
else:
   import mfem.ser as mfem

from petram.phys.vtable import VtableElement, Vtable      
data =  (('jext', VtableElement('jext', type='complex',
                             guilabel = 'External J',
                             suffix =('x', 'y', 'z'),
                             default = [0,0,0],
                             tip = "volumetric external current" )),)

from petram.phys.coefficient import VCoeff
#from petram.phys.coefficient import PyComplexVectorSliceCoefficient as ComplexVectorSlice
   
def jwJ_Coeff(exprs, ind_vars, l, g, omega):
    # iomega x Jext
    fac =  1j * omega
    return VCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)

def domain_constraints():
   return [EM1D_ExtJ]
 
class EM1D_ExtJ(EM1D_Domain):
    is_secondary_condition = True  # does not count this class for persing "remaining"
    has_3rd_panel = False

    vt  = Vtable(data)
    
    def has_lf_contribution(self, kfes = 0):
        if kfes == 0: return True
        if kfes == 1: return True
        if kfes == 2: return True
        return False

    def add_lf_contribution(self, engine, b, real = True, kfes = 0):
        if kfes < 3:
            if real:       
                dprint1("Add LF contribution(real)" + str(self._sel_index))
            else:
                dprint1("Add LF contribution(imag)" + str(self._sel_index))
            freq, omega = self.get_root_phys().get_freq_omega()
            self.set_integrator_realimag_mode(real)

            jext = self.vt.make_value_or_expression(self)            
            jwJ = jwJ_Coeff(jext[0], self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns, omega)
            

            if kfes in [0, 1, 2]:
                #  j_slice = ComplexVectorSlice(jwJ, [kfes])
                j_slice = jwJ[kfes]
                self.add_integrator(engine, 'jext', j_slice,
                                    b.AddDomainIntegrator,
                                    mfem.DomainLFIntegrator)

        else:
           assert False, "should not come here"

               
            



        
        
        




'''
   PML

   * Linear PML
            [S(x)          ]
        S = [     S(y)     ]

        S  = 1 + (Sr + iSi)*((r -Lr)/Lpml)^p

        Lr = the coordinate of boundar between normal and PML
        Lpml = PML thickness
        Sr, Si = PML strech factor
        p = the order of PML stretch

      We replase epsilon and mu in the PML region as follow
         epsilonr -> G * epsilon * S^-1, 
         mur -> G * mur * S^-1,
      where G = det(S) S^-1. det(S) is S(x)*S(y)*S(z) for linear case

      CopyRight (c) 2020-  S. Shiraiwa
'''
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.phys_const import mu0, epsilon0
from petram.phys.pycomplex_coefficient import CC_Matrix
from petram.mfem_config import use_parallel
import numpy as np
import abc
from abc import ABC, abstractmethod

from petram.phys.phys_model import Phys, PhysCoefficient, MatrixPhysCoefficient
from petram.phys.em2d.em2d_base import EM2D_Bdry, EM2D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2D_PML')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('stretch', VtableElement('stretch', type='complex',
                                  guilabel='stretch',
                                  default=1.0,
                                  no_func=True,
                                  tip="real: shorten wavelength (>0), imag: damping (>0 forward, <0 backward)")),
        ('s_order', VtableElement('s_order', type='float',
                                  guilabel='order',
                                  default=1.0,
                                  no_func=True,
                                  tip="streach order")),)


class EM2D_PML(EM2D_Domain, ABC):
    @abstractmethod
    def make_PML_coeff(self):
        pass

    def get_parent_selection(self):
        return self.parent.sel_index


class LinearPML(CC_Matrix):
    def __init__(self, coeff, S, direction, ref_point, pml_width, order, inv):
        self.coeffs = coeff
        self.direction = direction
        self.ref_p = ref_point
        self.pml_width = pml_width
        self.S = S
        self.order = order
        self.return_inv = inv
        self.coeff = coeff
        super(LinearPML, self).__init__(3, 3)

    def eval(self, T, ip):
        ptx = T.Transform(ip)

        invS = self.Eval_invS(ptx)
        detS = self.Eval_detS(ptx)

        K_m = self.coeffs.eval(T, ip)

        detS_inv_S_x_inv_S = (invS*detS).dot(K_m).dot(invS)

        if self.return_inv:
            detS_inv_S_x_inv_S = np.linalg.inv(detS_inv_S_x_inv_S)

        return detS_inv_S_x_inv_S

    def Eval_S(self, x):
        ret = np.array([1+0j, 1., 1.])

        if self.direction[0]:
            ret[0] = 1 + self.S * \
                ((x[0] - self.ref_p[0])/self.pml_width[0])**self.order
        if self.direction[1]:
            ret[1] = 1 + self.S * \
                ((x[1] - self.ref_p[1])/self.pml_width[1])**self.order
        return ret

    def Eval_invS(self, x):
        ret = self.Eval_S(x)
        return np.diag((1+0j)/ret)

    def Eval_detS(self, x):
        ret = self.Eval_S(x)
        return ret[0]*ret[1]

def linear_pml_func1(ptx, K_m):
    Sd = np.array([1+0j, 1., 1.])

    if direction[0]:
        Sd[0] = 1 + S * ((ptx[0] - ref_p[0])/pml_width[0])**pml_order
    if direction[1]:
        Sd[1] = 1 + S * ((ptx[1] - ref_p[1])/pml_width[1])**pml_order

    invS = np.diag((1+0j)/Sd)
    detS = Sd[0]*Sd[1]
    K_m = np.ascontiguousarray(K_m)

    detS_inv_S_x_inv_S = (invS*detS).dot(K_m).dot(invS)

    return detS_inv_S_x_inv_S

def linear_pml_func2(ptx):
    Sd = np.array([1+0j, 1., 1.])

    if direction[0]:
        Sd[0] = 1 + S * ((ptx[0] - ref_p[0])/pml_width[0])**pml_order
    if direction[1]:
        Sd[1] = 1 + S * ((ptx[1] - ref_p[1])/pml_width[1])**pml_order

    invS = np.diag((1+0j)/Sd)
    detS = Sd[0]*Sd[1]

    detS_inv_S_x_inv_S = (invS*detS).dot(K_m).dot(invS)

    return detS_inv_S_x_inv_S


class EM2D_LinearPML(EM2D_PML):
    has_2nd_panel = False
    has_3rd_panel = False
    _has_4th_panel = False
    vt = Vtable(data)

    def attribute_set(self, v):
        v = super(EM2D_LinearPML, self).attribute_set(v)
        v["stretch_dir"] = [False, False]
        v["ref_point"] = ""
        return v

    def panel1_param(self):
        panels1 = [["Direction", self.stretch_dir, 36,
                    {'col': 3, 'labels': ['x', 'y']}],
                   ["Ref. point", self.ref_point, 0, {}],
                   ]
        panels2 = super(EM2D_LinearPML, self).panel1_param()

        panels = panels1 + panels2
        return panels

    def get_panel1_value(self):
        val = super(EM2D_LinearPML, self).get_panel1_value()
        val = [self.stretch_dir, self.ref_point] + val
        return val

    def import_panel1_value(self, v):
        super(EM2D_LinearPML, self).import_panel1_value(v[2:])
        self.ref_point = str(v[1])
        self.stretch_dir = [vv[1] for vv in v[0]]

    def panel1_tip(self):
        tip1 = ["PML direction", "reference point to measure the PML thickness"]
        tip2 = super(EM2D_LinearPML, self).panel1_tip()
        return tip1 + tip2

    def preprocess_params(self, engine):
        super(EM2D_LinearPML, self).preprocess_params(engine)
        dprint1("PML (stretch, order) ", self.stretch, self.s_order)

        base_mesh = engine.meshes[0]

        ec = base_mesh.extended_connectivity
        v2v = ec['vert2vert']

        idx = int(self.ref_point)
        self.ref_point_coord = base_mesh.GetVertexArray(v2v[idx]).copy()

        dprint1("PML reference point:", self.ref_point_coord)

        l = sum([ec['surf2line'][k] for k in self.get_parent_selection()], [])
        l = list(set(l))
        v = sum([ec['line2vert'][k] for k in l], [])
        v = list(set(v))

        ptx = np.vstack([base_mesh.GetVertexArray(v2v[idx]) for idx in v])

        self.pml_width = [0]*2
        for k in range(2):
            dd = ptx[:, k] - self.ref_point_coord[k]
            ii = np.argmax(np.abs(dd))
            self.pml_width[k] = dd[ii]

        dprint1("PML width ", self.pml_width)

    def make_PML_coeff(self, coeff):
        from petram.phys.numba_coefficient import NumbaCoefficient
        from petram.mfem_config import numba_debug
        from petram.phys.pycomplex_coefficient import (PyComplexConstant,
                                                       PyComplexMatrixConstant,)

        if isinstance(coeff, (NumbaCoefficient,
                              PyComplexConstant,
                              PyComplexMatrixConstant)):
            params = {"S": self.stretch,
                      "direction": np.array(self.stretch_dir),
                      "ref_p": np.array(self.ref_point_coord),
                      "pml_width": np.array(self.pml_width),
                      "pml_order": self.s_order, }

            if isinstance(coeff, NumbaCoefficient):
                dep = (coeff.mfem_numba_coeff,)
                coeff = mfem.jit.matrix(sdim=2,
                                        shape=(3, 3),
                                        complex=True,
                                        debug=numba_debug,
                                        dependency=dep,
                                        params=params,
                                        interface="simple")(linear_pml_func1)
            else:
                if isinstance(coeff, PyComplexConstant):
                    value = np.diag([coeff.value]*3)
                else:
                    value = coeff.value
                params["K_m"] = value
                coeff = mfem.jit.matrix(sdim=2,
                                        shape=(3, 3),
                                        complex=True,
                                        debug=numba_debug,
                                        params=params,
                                        interface="simple")(linear_pml_func2)

            coeff = NumbaCoefficient(coeff)
        else:
            coeff = LinearPML(coeff, self.stretch, self.stretch_dir,
                              self.ref_point_coord, self.pml_width, self.s_order,
                              False)
        return coeff
    '''
    def make_PML_invmu(self):
        coeff = LinearPML(coeff, self.stretch, self.stretch_dir,
                          self.ref_point_coord, self.pml_width, self.s_order,
                          True)
        return coeff

    def make_PML_sigma(self):
        coeff = LinearPML(coeff, self.stretch, self.stretch_dir,
                          self.ref_point_coord, self.pml_width, self.s_order,
                          False)
        return coeff
    '''



'''

  evaulate derivative of Nedelec element using
  DiscreteLinearOperator

'''
import mfem
import numpy as np

from petram.mfem_config import use_parallel
import petram.debug

dprint1, dprint2, dprint3 = petram.debug.init_dprints('eval_deriv')

if mfem.__version__.startswith('4.2'):
    isMFEM42 = True
else:
    isMFEM42 = False

if use_parallel:
    import mfem.par as mfem
    FiniteElementSpace = mfem.ParFiniteElementSpace
    DiscreteLinearOperator = mfem.ParDiscreteLinearOperator
    GridFunction = mfem.ParGridFunction
else:
    import mfem.ser as mfem
    FiniteElementSpace = mfem.FiniteElementSpace
    DiscreteLinearOperator = mfem.DiscreteLinearOperator
    GridFunction = mfem.GridFunction


def eval_curl(gfr, gfi=None):
    '''
    evaluate curl, gfr/gfi is supposed to be Et
    '''
    fes = gfr.FESpace()
    ordering = fes.GetOrdering()
    mesh = fes.GetMesh()
    vdim = 1
    sdim = mesh.SpaceDimension()
    p = fes.GetOrder(0)
    rt_coll = mfem.L2_FECollection(p - 1, sdim)

    rts = FiniteElementSpace(mesh, rt_coll, vdim, ordering)

    curl = DiscreteLinearOperator(fes, rts)
    itp = mfem.CurlInterpolator()
    curl.AddDomainInterpolator(itp)
    curl.Assemble()
    curl.Finalize()

    br = GridFunction(rts)
    curl.Mult(gfr, br)
    if gfi is not None:
        bi = GridFunction(rts)
        curl.Mult(gfi, bi)
    else:
        bi = None

    if isMFEM42:
        # this is meant for adjustment only for MFEM4.2.
        # it is confirmed not needed in 4.3 (and probably later version of 4.2)
        arr = br.GetDataArray()
        scale = np.zeros(len(arr))

        for ii in range(rts.GetNE()):
            trans = rts.GetElementTransformation(ii)
            dofs = rts.GetElementDofs(ii)
            scale[dofs] = trans.Weight()

        brr = br.GetDataArray()
        brr /= scale
        if bi is not None:
            bii = bi.GetDataArray()
            bii /= scale

    return br, bi, rts


def eval_grad(gfr, gfi=None):
    '''
    evaluate grad
    '''
    fes = gfr.FESpace()
    ordering = fes.GetOrdering()
    mesh = fes.GetMesh()
    vdim = 1
    sdim = mesh.SpaceDimension()
    p = fes.GetOrder(0)
    #rt_coll = mfem.L2_FECollection(p - 1, sdim)
    #rt_coll = mfem.ND_FECollection(p - 1, sdim)
    rt_coll = mfem.ND_FECollection(p, sdim)

    rts = FiniteElementSpace(mesh, rt_coll, vdim, ordering)

    grad = DiscreteLinearOperator(fes, rts)
    itp = mfem.GradientInterpolator()
    grad.AddDomainInterpolator(itp)
    grad.Assemble()
    grad.Finalize()

    br = GridFunction(rts)
    grad.Mult(gfr, br)
    if gfi is not None:
        bi = GridFunction(rts)
        grad.Mult(gfi, bi)
    else:
        bi = None
    # needs to return rts to prevent rts to be collected.
    return br, bi, rts



'''
   external current source
'''
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.coefficient import VCoeff
from petram.mfem_config import use_parallel
import numpy as np

from petram.phys.phys_model import VectorPhysCoefficient
from petram.phys.phys_model import PhysCoefficient
from petram.phys.em2d.em2d_base import EM2D_Bdry, EM2D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2D_extJ')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('jext', VtableElement('jext', type='complex',
                               guilabel='External J',
                               suffix=('x', 'y', 'z'),
                               default=[0, 0, 0],
                               tip="volumetric external current")),)


#from petram.phys.coefficient import PyComplexVectorSliceCoefficient as ComplexVectorSlice


def jwJ_Coeff(exprs, ind_vars, l, g, omega):
    # iomega x Jext
    fac = 1j * omega
    return VCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)

def domain_constraints():
   return [EM2D_ExtJ]

class EM2D_ExtJ(EM2D_Domain):
    is_secondary_condition = True  # does not count this class for persing "remaining"
    has_3rd_panel = False
    vt = Vtable(data)

    def has_lf_contribution(self, kfes=0):
        if kfes > 2:
            return False
        return True

    def add_lf_contribution(self, engine, b, real=True, kfes=0):
        if kfes < 2:
            if real:
                dprint1("Add LF contribution(real)" + str(self._sel_index))
            else:
                dprint1("Add LF contribution(imag)" + str(self._sel_index))
            freq, omega = self.get_root_phys().get_freq_omega()
            f_name = self.vt.make_value_or_expression(self)
            self.set_integrator_realimag_mode(real)

            jwJ = jwJ_Coeff(f_name[0], self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns, omega)

            if kfes == 0:
                #jwJ01 = ComplexVectorSlice(jwJ, [0,1])
                jwJ01 = jwJ[[0, 1]]

                self.add_integrator(engine, 'jext', jwJ01,
                                    b.AddDomainIntegrator,
                                    mfem.VectorFEDomainLFIntegrator)
            else:
                jwJ2 = jwJ[2]
                #jwJ2 = ComplexVectorSlice(jwJ, [2])
                self.add_integrator(engine, 'jext', jwJ2,
                                    b.AddDomainIntegrator,
                                    mfem.DomainLFIntegrator)

        else:
            assert False, "should not come here"



'''
   write_cdf: write netCDF file

       solset: solset
       battrs: list of boundary index

       curl : True/False if include curlE in file

       params : a list of callables
                other parameters evaluated on all x, y, z
                these should be namelist params.

   for cold plasma
        P, D, S, Br, Bz, Bt, Te, ne
        
'''   
from petram.utils import eval_sol
from petram.utils import eval_curl
from netCDF4 import Dataset
from numpy import pi, exp, abs, concatenate, dstack, array

def eval_solset(solsets, battrs, func, dim):
    ret = {}
    for battr in battrs:
        ret[battr] = []
        for m0, solr, soli in solsets:
             if soli is None:
                v,  cdata = func(solr, battr, dim)
                if v is None: continue
                ret[battr].append((v, cdata))
             else:
                v, cdata1 = func(solr, battr, dim)
                if v is None: continue
                v, cdata2 = func(soli, battr, dim)
                ret[battr].append((v, (cdata1 + 1j*cdata2)))
                #ret[battr].append((v, abs(cdata1 + 1j*cdata2)))
    return ret

def write_complex(rootgrp, name, value):
    ReE = rootgrp.createVariable("Re"+name, "f8", ("nele", "npts"))
    ImE = rootgrp.createVariable("Im"+name, "f8", ("nele", "npts"))
    ReE[:] = value.real
    ImE[:] = value.imag

def write_float(rootgrp, name, value):
    E = rootgrp.createVariable(name, "f8", ("nele", "npts"))
    E[:] = value

def create_cdf(filename, solsets, battrs):
    rootgrp = Dataset(filename, 'w', format='NETCDF3_64BIT')
    v  = eval_solset(solsets, battrs, eval_sol,  1)
    xy   = concatenate([concatenate([x[0] for x in v[i][:]]) for i in battrs])
    nele = rootgrp.createDimension('nele', xy.shape[0])
    ndim = rootgrp.createDimension('ndim', xy.shape[-1])
    npts = rootgrp.createDimension('npts', xy.shape[1])
    write_float(rootgrp, 'x', xy[:,:,0])
    write_float(rootgrp, 'y', xy[:,:,1])
    write_float(rootgrp, 'z', xy[:,:,2])
#    pts = rootgrp.createVariable("pts", "f8", ("nele", "npts", "ndim"))
#    pts[:] = xy
    rootgrp.close()

def add_sol_variable(filename, solsets, battrs, curl=False):
    rootgrp = Dataset(filename, 'a')    
    Ex  = eval_solset(solsets, battrs, eval_sol,  1)
    Ey  = eval_solset(solsets, battrs, eval_sol,  2)
    Ez  = eval_solset(solsets, battrs, eval_sol,  3)

    Exa  = concatenate([concatenate([x[1] for x in Ex[i][:]]) for i in battrs])
    Eya  = concatenate([concatenate([x[1] for x in Ey[i][:]]) for i in battrs])
    Eza  = concatenate([concatenate([x[1] for x in Ez[i][:]]) for i in battrs])

    write_complex(rootgrp, "Ex", Exa)
    write_complex(rootgrp, "Ey", Eya)
    write_complex(rootgrp, "Ez", Eza)

    if curl:
        cEx = eval_solset(solsets, battrs, eval_curl, 1)
        cEy = eval_solset(solsets, battrs, eval_curl, 2)
        cEz = eval_solset(solsets, battrs, eval_curl, 3)
        cExa = concatenate([concatenate([x[1] for x in cEx[i][:]]) for i in battrs])
        cEya = concatenate([concatenate([x[1] for x in cEy[i][:]]) for i in battrs])
        cEza = concatenate([concatenate([x[1] for x in cEz[i][:]]) for i in battrs])

        write_complex(rootgrp, "cEx", cExa)
        write_complex(rootgrp, "cEy", cEya)
        write_complex(rootgrp, "cEz", cEza)

def add_ns_variable(filename, namespace, name, complex=False):
    rootgrp = Dataset(filename, 'r')
    x = rootgrp.variables['x'][:]
    y = rootgrp.variables['y'][:]
    z = rootgrp.variables['z'][:]

    pts = dstack((x, y, z))
    rootgrp.close()
    pp = pts.reshape(-1, pts.shape[-1])
    f =  namespace[name]
    data = array([f(*p) for p in pp])
    data = data.reshape((pts.shape[0], pts.shape[1]))

    rootgrp = Dataset(filename, 'a')
    if complex:
        write_complex(rootgrp, name, data)
    else:
        write_float(rootgrp, name, data)
    rootgrp.close()

def example(solsets, ns, filename='icrf.nc', battrs = [9]):
    create_cdf('icrf.nc', solsets, battrs)
    add_sol_variable(filename, solsets, battrs, curl=True)

    add_ns_variable(filename, ns, 'P_rf', complex=True)
    add_ns_variable(filename, ns, 'S_rf', complex=True)
    add_ns_variable(filename, ns, 'D_rf', complex=True)

    add_ns_variable(filename, ns, 'ne')
    add_ns_variable(filename, ns, 'ni')
    add_ns_variable(filename, ns, 'nim')
    add_ns_variable(filename, ns, 'Te')
    add_ns_variable(filename, ns, 'Ti')
    add_ns_variable(filename, ns, 'Tim')

    add_ns_variable(filename, ns, 'Br')
    add_ns_variable(filename, ns, 'Bz')
    add_ns_variable(filename, ns, 'Bt')



   



'''
   divergence = 0 constraint

   this constraints add weak term, (grad u,  epsilon E), to the weak formulation.
   u is Lagurange multiplier, and descritized by H1 element.  

'''
import numpy as np
from scipy.sparse import csr_matrix, coo_matrix, lil_matrix, csc_matrix


from petram.model import Domain, Bdry, Pair
from petram.phys.phys_model  import Phys, VectorPhysCoefficient
import numpy as np

import petram.debug as debug
#from petram.helper.chypre_to_pymatrix import *
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_Div')

from petram.mfem_config import use_parallel
if use_parallel:
   import mfem.par as mfem
   from mpi4py import MPI
   num_proc = MPI.COMM_WORLD.size
   myid     = MPI.COMM_WORLD.rank
else:
   import mfem.ser as mfem

#from mfem.common.sparse_utils import eliminate_rows, eliminate_cols , sparsemat_to_scipycsr
class Arctan(mfem.PyCoefficient):
   def EvalValue(self, x):
       return  np.arctan2(x[0], x[2])

def domain_constraints():
   return [EM3D_Div]
    
class EM3D_Div(Domain, Phys):
    is_secondary_condition = True   
    has_essential = True
    def __init__(self, **kwargs):
        super(EM3D_Div, self).__init__(**kwargs)
        Phys.__init__(self)
        
    def attribute_set(self, v):
        super(EM3D_Div, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        v['psi0'] = 1.0
        v['essential_bdr'] = ''
        return v
    
    def panel1_param(self):
        return [["psi",  float(self.psi0),  300, {}],
                ["essential_bdr",  self.essential_bdr,  0, {}],]
                
    def get_panel1_value(self):
        return (float(self.psi0),
                self.essential_bdr)

    def import_panel1_value(self, v):
        self.psi0 = float(v[0])
        self.essential_bdr = str(v[1])

       #def has_extra_DoF(self):
    #    return True

    def get_essential_idx(self, idx):
        if idx == 0: 
            return []
        else:
            if self.essential_bdr.strip() == '': return []
            bdr = [int(x) for x in self.essential_bdr.strip().split(',')]
            return bdr

    def has_mixed_contribution(self):
        return True

    def get_mixedbf_loc(self):
        '''
        r, c, and flag of MixedBilinearForm
           flag1 : take transpose
           flag2 : take conj
        '''
        return [(1, 0, -1, -1), (1, 0, 1, 1)]

    ''' 
    def postprocess_extra(self, sol, flag, sol_extra):
        name = 'psi_div'
        sol_extra[name] = np.array(sol)
    '''

    def add_mix_contribution(self, engine, mbf, r, c, is_trans, real = True):
        dprint1("Add Mixed contribution: " + str((r, c, is_trans, real)))
        itg =  mfem.MixedVectorWeakDivergenceIntegrator
        domains = [engine.find_domain_by_index(self.get_root_phys(), x,
                                               check_enabled = True)
                   for x in self._sel_index]
        
        self.set_integrator_realimag_mode(real)        
        for dom, idx in zip(domains, self._sel_index):
            coeff1, coeff2, coeff3 = dom.get_coeffs()
            
            if dom.has_pml():
               coeff1 = dom.make_PML_epsilon(coeff1)
               coeff3 = dom.make_PML_sigma(coeff3)
           
            self.add_integrator(engine, 'epsilonr', coeff1,
                                mbf.AddDomainIntegrator,
                                itg, idx=[idx], vt  = dom.vt, transpose=is_trans)
            self.add_integrator(engine, 'sigma', coeff3,
                                mbf.AddDomainIntegrator,
                                itg, idx=[idx], vt  = dom.vt, transpose=is_trans)
            
    def apply_essential(self, engine, gf, real = False, kfes = 0):
        if kfes == 0: return
        if real:       
            dprint1("Apply Ess.(real)" + str(self._sel_index))
            coeff = mfem.ConstantCoefficient(float(self.psi0))
        else:
            return
            #dprint1("Apply Ess.(imag)" + str(self._sel_index))
            #coeff = mfem.ConstantCoefficient(complex(self.psi0).imag)

        coeff = self.restrict_coeff(coeff, engine)
        mesh = engine.get_mesh(mm = self)        
        ibdr = mesh.bdr_attributes.ToList()
        bdr_attr = [0]*mesh.bdr_attributes.Max()

        ess_bdr = self.get_essential_idx(1)      
        dprint1("Essential boundaries", ess_bdr)
 
        for idx in ess_bdr:
            bdr_attr[idx-1] = 1
        gf.ProjectBdrCoefficient(coeff, mfem.intArray(bdr_attr))

    def add_bf_contribution(self, engine, a, real = True, kfes=0):
        '''
        no dialoganl element except for essentail DoF
        '''
        pass
        



import traceback
import numpy as np

from petram.model import Domain, Bdry, Pair
from petram.phys.phys_model import Phys, PhysModule, VectorPhysCoefficient, PhysCoefficient

# define variable for this BC.
from petram.phys.vtable import VtableElement, Vtable
data =  (('Einit', VtableElement('Einit', type='float',
                                  guilabel = 'E(init)',
                                  suffix =('r', 'phi', 'z'),
                                  default = np.array([0,0,0]), 
                                  tip = "initial_E",
                                  chkbox = True)),)
class Einit_p(VectorPhysCoefficient):
   def EvalValue(self, x):
       v = super(Einit_p, self).EvalValue(x)
       v = np.array((v[0], v[2]))
       if self.real:  val = v.real
       else: val =  v.imag
       return val
class rEinit_t(PhysCoefficient):
   def EvalValue(self, x):
       v = super(rEinit_t, self).EvalValue(x)
       v = v[1]*x[0]
       if self.real:  val = v.real
       else: val =  v.imag
       return val
    

class EM1D_Domain(Domain, Phys):
    has_3rd_panel = True    
    vt3  = Vtable(data)   
    def __init__(self, **kwargs):
        super(EM1D_Domain, self).__init__(**kwargs)
        Phys.__init__(self)
        
    def attribute_set(self, v):
        super(EM1D_Domain, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v
    
    def get_init_coeff(self, engine, real=True, kfes=0):
        if kfes > 2: return
        if not self.use_Einit: return
        
        f_name = self.vt3.make_value_or_expression(self)
        if kfes == 0:
            coeff = Einit_p(2, f_name[0],
                       self.get_root_phys().ind_vars,
                       self._local_ns, self._global_ns,
                       real = real)
            return self.restrict_coeff(coeff, engine, vec = True)            
        else:
            coeff = rEinit_t(f_name[0],
                       self.get_root_phys().ind_vars,
                       self._local_ns, self._global_ns,
                       real = real)
            return self.restrict_coeff(coeff, engine)                        


class EM1D_Bdry(Bdry, Phys):
    has_3rd_panel = True        
    vt3  = Vtable(data)   
    def __init__(self, **kwargs):
        super(EM1D_Bdry, self).__init__(**kwargs)
        Phys.__init__(self)
        
    def attribute_set(self, v):
        super(EM1D_Bdry, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    def get_init_coeff(self, engine, real=True, kfes=0):
        if kfes > 2: return
        if not self.use_Einit: return
        
        f_name = self.vt3.make_value_or_expression(self)
        if kfes == 0:
            coeff = Einit_p(2, f_name[0],
                       self.get_root_phys().ind_vars,
                       self._local_ns, self._global_ns,
                       real = real)
            return self.restrict_coeff(coeff, engine, vec = True)            
        else:
            coeff = rEinit_t(f_name[0],
                       self.get_root_phys().ind_vars,
                       self._local_ns, self._global_ns,
                       real = real)
            return self.restrict_coeff(coeff, engine)                                    


     




import numpy as np

from petram.mfem_config import use_parallel
import petram.debug

dprint1, dprint2, dprint3 = petram.debug.init_dprints('eval_deriv')

if use_parallel:
    import mfem.par as mfem
    FiniteElementSpace = mfem.ParFiniteElementSpace
    DiscreteLinearOperator = mfem.ParDiscreteLinearOperator
    GridFunction = mfem.ParGridFunction
else:
    import mfem.ser as mfem
    FiniteElementSpace = mfem.FiniteElementSpace
    DiscreteLinearOperator = mfem.DiscreteLinearOperator
    GridFunction = mfem.GridFunction

def eval_grad(gfr, gfi=None):
    '''
    evaluate grad
    '''
    fes = gfr.FESpace()
    ordering = fes.GetOrdering()
    mesh = fes.GetMesh()
    vdim = 1
    sdim = mesh.SpaceDimension()
    p = fes.GetOrder(0)
    l2_coll = mfem.L2_FECollection(p - 1, sdim)

    rts = FiniteElementSpace(mesh, l2_coll, vdim, ordering)

    grad = DiscreteLinearOperator(fes, rts)
    itp = mfem.GradientInterpolator()
    grad.AddDomainInterpolator(itp)
    grad.Assemble()
    grad.Finalize()

    br = GridFunction(rts)
    grad.Mult(gfr, br)
    if gfi is not None:
        bi = GridFunction(rts)
        grad.Mult(gfi, bi)
    else:
        bi = None
    # needs to return rts to prevent rts to be collected.
    return br, bi, rts



from petram.phys.em2d.em2d_base import EM2D_Bdry, EM2D_Domain
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2D_PEC')

from petram.mfem_config import use_parallel
if use_parallel:
   import mfem.par as mfem
else:
   import mfem.ser as mfem

from petram.phys.vtable import VtableElement, Vtable   
data =  (('label1', VtableElement(None, 
                                  guilabel = 'Perfect Electric Conductor',
                                  default =   "Et = 0",
                                  tip = "Essential Homogenous BC" )),)
def bdry_constraints():
   return [EM2D_PEC]

class EM2D_PEC(EM2D_Bdry):
    has_essential = True
    nlterms = []
    vt  = Vtable(data)          
    
    def get_essential_idx(self, kfes):
        if kfes == 0:
            return self._sel_index
        elif kfes == 1:
            return self._sel_index            
        else:
            return []

    def apply_essential(self, engine, gf, real = False, kfes = 0):
        if kfes > 1: return
        if real:       
            dprint1("Apply Ess.(real)" + str(self._sel_index))
        else:
            dprint1("Apply Ess.(imag)" + str(self._sel_index))
            
        Erphiz = self.vt.make_value_or_expression(self)              
        mesh = engine.get_mesh(mm = self)        
        ibdr = mesh.bdr_attributes.ToList()
        bdr_attr = [0]*mesh.bdr_attributes.Max()
        for idx in self._sel_index:
            bdr_attr[idx-1] = 1

        if kfes == 0:
            coeff1 = mfem.VectorArrayCoefficient(2)
            coeff1.Set(0, mfem.ConstantCoefficient(0.0))
            coeff1.Set(1, mfem.ConstantCoefficient(0.0))            
            gf.ProjectBdrCoefficientTangent(coeff1,
                                            mfem.intArray(bdr_attr))
        elif kfes == 1:
            coeff1 = mfem.ConstantCoefficient(0.0)
            gf.ProjectBdrCoefficient(coeff1,
                                     mfem.intArray(bdr_attr))

        





'''
   external current source
'''
from petram.phys.phys_model import PhysCoefficient
from petram.phys.phys_model import VectorPhysCoefficient
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.coefficient import VCoeff, SCoeff
from petram.mfem_config import use_parallel
import numpy as np

from petram.phys.em2da.em2da_base import EM2Da_Bdry, EM2Da_Domain
from petram.phys.numba_coefficient import (func_to_numba_coeff_scalar,
                                           func_to_numba_coeff_vector,
                                           func_to_numba_coeff_matrix)

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2Da_extJ')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('jext', VtableElement('jext', type='complex',
                               guilabel='External J',
                               suffix=('r', 'phi', 'z'),
                               default=[0, 0, 0],
                               tip="volumetric external current")),)


'''
class rJext_p(VectorPhysCoefficient): # i \omega r Jext_t
   def __init__(self, *args, **kwargs):
       self.omega = kwargs.pop('omega', 1.0)
       super(rJext_p, self).__init__(*args, **kwargs)
   def EvalValue(self, x):
       v = super(rJext_p, self).EvalValue(x)
       v = np.array((v[0], v[2]))
       v = 1j * self.omega * v * x[0]
       if self.real:  return v.real
       else: return v.imag
       
class Jext_t(PhysCoefficient):  # i \omega Jext_phi
   def __init__(self, *args, **kwargs):
       self.omega = kwargs.pop('omega', 1.0)
       super(Jext_t, self).__init__(*args, **kwargs)
   def EvalValue(self, x):
       v = super(Jext_t, self).EvalValue(x)
       v = 1j * self.omega * v[1]
       if self.real:  return v.real
       else: return v.imag
'''


def JextCoeff(exprs, ind_vars, l, g, omega):
    fac = 1j * omega
    coeff = VCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)
    return coeff


def domain_constraints():
    return [EM2Da_ExtJ]


class EM2Da_ExtJ(EM2Da_Domain):
    is_secondary_condition = True  # does not count this class for persing "remaining"
    has_3rd_panel = False
    vt = Vtable(data)

    def has_lf_contribution(self, kfes=0):
        if kfes > 2:
            return False
        return True

    def add_lf_contribution(self, engine, b, real=True, kfes=0):
        if kfes < 2:
            if real:
                dprint1("Add LF contribution(real)" + str(self._sel_index))
            else:
                dprint1("Add LF contribution(imag)" + str(self._sel_index))
            freq, omega = self.get_root_phys().get_freq_omega()
            f_name = self.vt.make_value_or_expression(self)
            ind_vars = self.get_root_phys().ind_vars

            jext = JextCoeff(f_name[0], ind_vars, self._local_ns, self._global_ns,
                             omega)

            def rjext_p(ptx, jext):
                return np.array((jext[0]*ptx[0], jext[2]*ptx[0]), dtype=np.complex128)

            def jext_t(ptx, jext):
                return jext[1]

            if kfes == 0:
                '''
                coeff1 = rJext_p(2, f_name[0],  self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            real = real, omega = omega)
                '''
                coeff1 = func_to_numba_coeff_vector(rjext_p,
                                                    shape=(2, ),
                                                    complex=True,
                                                    dependency=(jext,))

                self.add_integrator(engine, 'jext', coeff1,
                                    b.AddDomainIntegrator,
                                    mfem.VectorFEDomainLFIntegrator)
            else:
                '''
                coeff1 = Jext_t(f_name[0],  self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            real = real, omega = omega)
                '''
                coeff1 = func_to_numba_coeff_scalar(jext_t, complex=True,
                                                    dependency=(jext,))
                self.add_integrator(engine, 'jext', coeff1,
                                    b.AddDomainIntegrator,
                                    mfem.DomainLFIntegrator)

        else:
            assert False, "should not come here"



'''
   Surface Impedance

   On external surface    Z(w) n \times H = n \times n \times E    (1)

   (note)
      Robbin BC is 
          n \times 1/mu curl E + \gamma  n \times n \times E = Q   (2) 

      Since 
            curl E = -dB/dt (Faraday's law)
            B = mu H

      with exp(- j \omega t), (2) becomes 
          j \omega  n \times H + \gamma  n \times n \times E = Q

      Note, we don't consider fero-magnetic material where mu changes
      with hysteresis.

      Therefore we use in (1)
          \gamma = j \omega / Z(\omega) 
          Q = 0
      and, the surface integral 
          - \gamma \int F \dot (n \times n \times E) d\Omega 
                 = \gamma \int F \dot E d\Omega 

      is implemented using mfem::VectorFEMassIntegrator.

      We implement two interface.

      1) Using \sigma, \epsilon, \mu, we define the impedance as

          Z(\omega) = sqrt((j \omega \mu))/(\sigam + j\omega \epsilon))

      or directrly specify Z(\omega).

      As a default value, we use 

         \sigma = 1./1.8e-8 = 5.6e7 (Cu at room temperatrue)
         \epsilonr = 1.0
         \mur = 1.0

      CopyRight (c) 2020-  S. Shiraiwa
''' 
import numpy as np

from petram.phys.phys_model  import Phys, PhysCoefficient
from petram.phys.em2d.em2d_base import EM2D_Bdry, EM2D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2D_Z')

from petram.mfem_config import use_parallel
if use_parallel:
   import mfem.par as mfem
else:
   import mfem.ser as mfem

from petram.phys.vtable import VtableElement, Vtable

data1 =  (('impedance', VtableElement('impedance', type='complex',
                                       guilabel = 'Z(w)',
                                       default = 1.0,
                                       tip = "impedance " )),)
data2 = (('epsilonr_bnd', VtableElement('epsilonr_bnd', type='complex',
                                    guilabel = 'epsilonr(bnd)',
                                    default = 1.0, 
                                    tip = "boundary relative permittivity" )),                                    
         ('mur_bnd', VtableElement('mur_bnd', type='complex',
                                guilabel = 'mur',
                                default = 1.0, 
                                tip = "boundary relative permeability" )),
         ('sigma_bnd', VtableElement('sigma_bnd', type='complex',
                                 guilabel = 'sigma',
                                 default = 5.6e7,
                                 tip = "boundary condutivey" )),)


vt1  = Vtable(data1)
vt2  = Vtable(data2)

data3 = (('Z_param', VtableElement('Z_param',
                                   type = 'selectable',
                                   guilabel = '',
                                   choices= ('Impedance', 'e/m/s'),
                                   vtables = (vt1, vt2),
                                   default = 'Impedance')),)

from petram.phys.coefficient import SCoeff
from petram.phys.phys_const import mu0, epsilon0

class ImpedanceByEMS(mfem.PyCoefficient):
   def __init__(self, ems,  ind_vars, l, g, omega, real):
       self.e = [SCoeff([ems[0]], ind_vars, l, g, real=True),
                 SCoeff([ems[0]], ind_vars, l, g, real=False)]
       self.m = [SCoeff([ems[1]], ind_vars, l, g, real=True),
                 SCoeff([ems[1]], ind_vars, l, g, real=False)]
       self.s = [SCoeff([ems[2]], ind_vars, l, g, real=True),
                 SCoeff([ems[2]], ind_vars, l, g, real=False)]

       self.omega = omega
       self.real = real
       mfem.PyCoefficient.__init__(self)

   def Eval(self, T, ip):

       e = self.e[0].Eval(T, ip) + 1j* self.e[1].Eval(T, ip)
       m = self.m[0].Eval(T, ip) + 1j* self.m[1].Eval(T, ip)
       s = self.s[0].Eval(T, ip) + 1j* self.s[1].Eval(T, ip)

       omega = self.omega
       z = np.sqrt((1j*omega*mu0*m)/(s + 1j*omega*e*epsilon0))
       gamma = -1j*omega/z
       if self.real:
           return gamma.real
       else:
           return gamma.imag

class ImpedanceByZ(mfem.PyCoefficient):
   def __init__(self, z,  ind_vars, l, g, omega, real):
       self.z = [SCoeff([z[0]], ind_vars, l, g, real=True),
                 SCoeff([z[0]], ind_vars, l, g, real=False)]

       self.omega = omega
       self.real = real
       mfem.PyCoefficient.__init__(self)

   def Eval(self, T, ip):
       z = self.z[0].Eval(T, ip) + 1j* self.z[1].Eval(T, ip)
       omega = self.omega
       gamma = -1j*omega/z
       if self.real:
           return gamma.real
       else:
           return gamma.imag          

def bdry_constraints():
   return [EM2D_Impedance]
        
class EM2D_Impedance(EM2D_Bdry):
    is_essential = False
    vt  = Vtable(data3)

    def has_bf_contribution(self, kfes=0):
        return True
    
    def add_bf_contribution(self, engine, b, real=True, kfes=0):
        if real:       
            dprint1("Add BF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add BF contribution(imag)" + str(self._sel_index))

        freq, omega = self.get_root_phys().get_freq_omega()        
        mode, parameters = self.vt.make_value_or_expression(self)[0]

        ind_vars = self.get_root_phys().ind_vars
        l = self._local_ns
        g = self._global_ns
        
        if mode == 'e/m/s':
            er, mr, s = parameters
            try:
                z = np.sqrt((1j*omega*mu0*mr)/(s + 1j*omega*er*epsilon0))
                gamma = -1j*omega/z
                coeff = SCoeff([gamma], ind_vars, l, g, real=real)
                #assert False, "cause error"
            except:
                #import traceback
                #traceback.print_exc()
                coeff = ImpedanceByEMS(parameters, ind_vars, l, g, omega, real)

        else:
            z = parameters[0]
            try:
                gamma = -1j*omega/z
                coeff = SCoeff([gamma], ind_vars, l, g, real=real)
                dprint1("Impedance", z)
                #assert False, "cause error"
            except:
                #import traceback
                #traceback.print_exc()
                coeff = ImpedanceByZ(parameters, ind_vars, l, g, omega, real)
                dprint1("Impedance", parameters)                              

        if kfes == 0:
            self.add_integrator(engine, 'impedance1', coeff,
                                b.AddBoundaryIntegrator,
                                mfem.VectorFEMassIntegrator)
        else:
            self.add_integrator(engine, 'impedance2', coeff,
                                b.AddBoundaryIntegrator,
                                mfem.MassIntegrator)
           
        '''
        coeff1 = self.restrict_coeff(coeff1, engine, vec = True)
        b.AddBoundaryIntegrator(mfem.VectorFEDomainLFIntegrator(coeff1))
        '''
       



from __future__ import print_function

import numpy as np
from scipy.sparse import csr_matrix, coo_matrix, lil_matrix

from petram.model import Pair, Bdry
from petram.phys.phys_model  import Phys

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2D_Floquet')

from petram.mfem_config import use_parallel

if use_parallel:
   from mfem.common.parcsr_extra import ToHypreParCSR, get_row_partitioning
   import mfem.par as mfem
   from mpi4py import MPI
   num_proc = MPI.COMM_WORLD.size
   myid     = MPI.COMM_WORLD.rank
   from mfem.common.mpi_debug import nicePrint
else:
   import mfem.ser as mfem
   num_proc = 1
   myid = 0

'''
   Map DoF from src surface (s1) to dst surface (s2)
   
   s1 and s2 should be plain.

   axis of rotation from s1 to s2 should be perpdicular to
   normal vectors of s1 and s2.

   Twist is not considered?

   For a fineite (non 0, non 180) complex phase difference
   compulex conjugate is returned, to force Lagrange multiplier
   real
'''
def pair_constraints():
   return [EM2D_Floquet]

class EM2D_Floquet(Pair,Bdry, Phys):
    def __init__(self,  **kwargs):
        super(EM2D_Floquet, self).__init__(**kwargs)
        Phys.__init__(self)
        
    def attribute_set(self, v):
        v['dphase'] = 0.0
        v['dphase_txt'] = '0.0'
        v['map_to_u'] = "x"
        v['tol_txt'] = "1e-4"
        v['tol'] = 1e-4
        v['use_multiplier'] = False
        super(EM2D_Floquet, self).attribute_set(v)
        return v
        
    def panel1_param(self):
        return [['u mapping ',  self.map_to_u, 0, {}],
                self.make_phys_param_panel('phase(deg)',  self.dphase_txt),
                self.make_phys_param_panel('tol.',  self.tol_txt),
                ["", "E_dst = exp(i*angle)*E_src" ,2, None],]     
#                ["use Lagrange multiplier",   self.use_multiplier,  3, {"text":""}],]     

    def get_panel1_value(self):
        return (self.map_to_u, self.dphase_txt,
                self.tol_txt, None, self.use_multiplier)

    def import_panel1_value(self, v):
        self.map_to_u = v[0]
        self.dphase_txt  = str(v[1])
        self.tol_txt  = str(v[2])
        #self.use_multiplier  = bool(v[4])
        self.use_multiplier  = False
        
    def preprocess_params(self, engine):
        ### find normal (outward) vector...
        map_to_u = ['def map_to_u(xy):',
                    '    import numpy as np',
                    '    x = xy[0]; y=xy[1]',
                    '    return '+self.map_to_u]

        self.func_u = map_to_u

        dprint1("u(x, y)")
        dprint1(self.func_u)

        try:
            exec('\n'.join(self.func_u))
        except:
            raise ValueError("Cannot complie mappling rule")

        try:
            self.dphase = float(self.eval_phys_expr(str(self.dphase_txt), 'dphase')[0])
            self.tol = float(self.eval_phys_expr(str(self.tol_txt), 'tol')[0])
        except:
            import traceback
            traceback.print_exc()
            raise ValueError("Cannot evaluate dphase/tolerance to float number")

    def has_extra_DoF(self, kfes = 0):
        return False

    def has_interpolation_contribution(self, kfes = 0):
        #if kfes != 0: return False
        if self.use_multiplier: return False
        return True

    def add_interpolation_contribution(self, engine, ess_tdof=None, kfes=0):
        dprint1("Add interpolation contribution(real)" + str(self._sel_index))
        dprint1("kfes = ", str(kfes))        
        dprint1("dphase = ", str(self.dphase))        



        trans1= ['def trans1(xy):',
                 '    import numpy as np',
                 '    x = xy[0]; y=xy[1]',
                 '    return np.array(['+self.map_to_u +'])']
        
        lns = {}
        exec('\n'.join(self.func_u), self._global_ns, lns) # this defines map_to_u
        exec('\n'.join(trans1), self._global_ns, lns)      # this defines trans1
        
        map_to_u = lns['map_to_u']
        trans1 = lns['trans1']
        
        tdof = [] if ess_tdof is None else ess_tdof
        src = self._src_index
        dst = [x for x in self._sel_index if not x in self._src_index]

        fes = engine.get_fes(self.get_root_phys(), kfes)

        #old version
        '''
        from petram.helper.dof_mapping_matrix import dof_mapping_matrix
        M, r, c = dof_mapping_matrix(src,  dst,  fes, ess_tdof, 
                                     engine, self.dphase,
                                     map_to_u = map_to_u, map_to_v = map_to_v,
                                     smap_to_u = map_to_u, smap_to_v = map_to_v,
                                     tol = self.tol)
        #nicePrint(M.GetRowPartArray())
        #nicePrint(M.GetColPartArray())
        #nicePrint(M.shape)
        '''
        from petram.helper.dof_map import projection_matrix
        M, r, c = projection_matrix(src, dst, fes, ess_tdof, fes2=fes,
                                    trans1 = trans1, dphase = self.dphase,
                                    tol = self.tol, mode='edge')
        return M, r, c
        
    



'''
   Essential BC
'''
from petram.phys.coefficient import VCoeff
from petram.phys.vtable import VtableElement, Vtable
from petram.mfem_config import use_parallel
import numpy as np

from petram.model import Bdry
from petram.phys.phys_model import Phys, VectorPhysCoefficient, PhysCoefficient
from petram.phys.em2d.em2d_base import EM2D_Bdry

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2D_E')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('E', VtableElement('E', type='complex',
                            guilabel='Electric field',
                            suffix=('x', 'y', 'z'),
                            default=np.array([0., 0., 0.]),
                            tip="essential BC")),)

'''
class Exy(VectorPhysCoefficient):
   def EvalValue(self, x):
       v = super(Exy, self).EvalValue(x)
       v = np.array((v[0], v[1]))
       if self.real:  return v.real
       else: return v.imag
   
class Ez(PhysCoefficient):
   def __init__(self, *args, **kwargs):
       #kwargs['isArray'] = True
       PhysCoefficient.__init__(self, *args, **kwargs)
   def EvalValue(self, x):
       v = super(Ez, self).EvalValue(x)
       v = v[2]
       if self.real:  return v.real
       else: return v.imag
'''
def bdry_constraints():
   return [EM2D_E]

class EM2D_E(EM2D_Bdry):
    has_essential = True
    vt = Vtable(data)

    def get_essential_idx(self, kfes):
        if kfes > 2:
            return
        return self._sel_index

    def apply_essential(self, engine, gf, real=False, kfes=0):
        if kfes > 1:
            return
        if real:
            dprint1("Apply Ess.(real)" + str(self._sel_index))
        else:
            dprint1("Apply Ess.(imag)" + str(self._sel_index))

        Exyz = self.vt.make_value_or_expression(self)[0]
        mesh = engine.get_mesh(mm=self)
        ibdr = mesh.bdr_attributes.ToList()
        bdr_attr = [0]*mesh.bdr_attributes.Max()
        for idx in self._sel_index:
            bdr_attr[idx-1] = 1

        ind_vars = self.get_root_phys().ind_vars
        l = self._local_ns
        g = self._global_ns

        coeff1 = VCoeff(3, Exyz, ind_vars, l, g, return_complex=True)

        if kfes == 0:
            '''
            coeff1 = Exy(2, Exyz,
                        self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        real = real)
            '''
            coeff1 = coeff1[[0, 1]]
            coeff1 = coeff1.get_realimag_coefficient(real)
            gf.ProjectBdrCoefficientTangent(coeff1,
                                            mfem.intArray(bdr_attr))
        elif kfes == 1:
            '''
            coeff1 = Ez(Exyz, self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        real = real)
            '''
            coeff1 = coeff1[2]
            coeff1 = coeff1.get_realimag_coefficient(real)
            gf.ProjectBdrCoefficient(coeff1,
                                     mfem.intArray(bdr_attr))



'''`
   Essential BC
'''
import numpy as np

from petram.model import Bdry
from petram.phys.phys_model  import Phys, PhysCoefficient
from petram.phys.em1d.em1d_base import EM1D_Bdry, EM1D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM1D_E')

from petram.mfem_config import use_parallel
if use_parallel:
   import mfem.par as mfem
else:
   import mfem.ser as mfem

from petram.phys.vtable import VtableElement, Vtable      
data =  (('E', VtableElement('E', type='complex',
                             guilabel = 'Ey and Ez',
                             suffix =('y','z'),
                             default = np.array([0, 0, ]),
                             tip = "essential BC" )),)

class Et(PhysCoefficient):
   def __init__(self, *args, **kwargs):
       #kwargs['isArray'] = True
       self.Eyz_idx = kwargs.pop('Eyz_idx')
       PhysCoefficient.__init__(self, *args, **kwargs)
   def EvalValue(self, x):
       v = super(Et, self).EvalValue(x)
       v = v[self.Eyz_idx]
       if self.real:  return v.real
       else: return v.imag
       
def bdry_constraints():
   return [EM1D_E]
   
class EM1D_E(EM1D_Bdry):
    has_essential = True
    vt  = Vtable(data)
    
    def get_essential_idx(self, kfes):
        if kfes > 3: return
        return self._sel_index

    def apply_essential(self, engine, gf, real = False, kfes = 0):
        if kfes == 0: return
        if real:       
            dprint1("Apply Ess.(real)" + str(self._sel_index))
        else:
            dprint1("Apply Ess.(imag)" + str(self._sel_index))
            
        Eyz = self.vt.make_value_or_expression(self)              
        mesh = engine.get_mesh(mm = self)        
        ibdr = mesh.bdr_attributes.ToList()
        bdr_attr = [0]*mesh.bdr_attributes.Max()
        for idx in self._sel_index:
            bdr_attr[idx-1] = 1

        if kfes == 1:
            coeff1 = Et(Eyz,
                        self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        real = real, Eyz_idx=0)
            gf.ProjectBdrCoefficient(coeff1,
                                            mfem.intArray(bdr_attr))
        elif kfes == 2:
            coeff1 = Et(Eyz,
                        self.get_root_phys().ind_vars,
                        self._local_ns, self._global_ns,
                        real = real, Eyz_idx=1)
            gf.ProjectBdrCoefficient(coeff1,
                                     mfem.intArray(bdr_attr))






        
        
        




'''
   Vacuum region:
      However, can have arbitrary scalar epsilon_r, mu_r, sigma


'''
import numpy as np

from petram.phys.phys_const import mu0, epsilon0

from petram.phys.phys_model  import PhysCoefficient, VectorPhysCoefficient
from petram.phys.phys_model  import MatrixPhysCoefficient, Coefficient_Evaluator
from petram.phys.em2da.em2da_base import EM2Da_Bdry, EM2Da_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2Da_Anisotropic')

from petram.mfem_config import use_parallel
if use_parallel:
   import mfem.par as mfem
else:
   import mfem.ser as mfem

from petram.phys.vtable import VtableElement, Vtable
data =  (('epsilonr', VtableElement('epsilonr', type='complex',
                                     guilabel = 'epsilonr',
                                     suffix =[('r', 'phi', 'z'), ('r', 'phi', 'z')],
                                     default = np.eye(3, 3),
                                     tip = "relative permittivity" )),
         ('mur', VtableElement('mur', type='complex',
                                     guilabel = 'mur',
                                     default = 1.0,
                                     tip = "relative permeability" )),
         ('sigma', VtableElement('sigma', type='complex',
                                     guilabel = 'sigma',
                                     suffix =[('r', 'phi', 'z'), ('r', 'phi', 'z')],
                                     default = np.zeros((3, 3)),
                                     tip = "contuctivity" )),
         ('t_mode', VtableElement('t_mode', type='float',
                                     guilabel = 'm',
                                     default = 0.0,
                                     tip = "mode number" )),)

'''
Expansion of matrix is as follows

               [e_rz  e_12 ][Erz ]
[Wrz, Wphx] =  [           ][    ] = Wrz e_rz Erz + Wrz e_12 Ephi
               [e_21  e_phi][Ephx]

                                + Wphi e_21 Erz + Wphi*e_phi*Ephi


  Erz = Er e_r + Ez e_z
  Ephx = rho Ephi
'''
class M_RZ(MatrixPhysCoefficient):
    def __init__(self, *args, **kwargs):
        self.omega = kwargs.pop('omega', 1.0)
        self.cnorm = kwargs.pop('cnorm')
        super(M_RZ, self).__init__(*args, **kwargs)
    def EvalValue(self, x):
        val = Coefficient_Evaluator.EvalValue(self, x).reshape(3,3)
        val = self.apply_factor(x, val)
        return val[[0,2], :][:, [0, 2]]

class M_12(VectorPhysCoefficient):
    def __init__(self, *args, **kwargs):
        self.omega = kwargs.pop('omega', 1.0)
        self.cnorm = kwargs.pop('cnorm')
        super(M_12, self).__init__(*args, **kwargs)
    def EvalValue(self, x):
        val = Coefficient_Evaluator.EvalValue(self, x).reshape(3,3)
        val = self.apply_factor(x, val)
        return val[[0, 2], [1]]

class M_21(VectorPhysCoefficient):
    def __init__(self, *args, **kwargs):
        self.omega = kwargs.pop('omega', 1.0)
        self.cnorm = kwargs.pop('cnorm')
        super(M_21, self).__init__(*args, **kwargs)
    def EvalValue(self, x):
        val = Coefficient_Evaluator.EvalValue(self, x).reshape(3,3)
        val = self.apply_factor(x, val)
        return val[[1], [0,2]]

class M_PHI(PhysCoefficient):
    def __init__(self, *args, **kwargs):
        kwargs['isArray'] = True
        self.omega = kwargs.pop('omega', 1.0)
        self.cnorm = kwargs.pop('cnorm')
        super(M_PHI, self).__init__(*args, **kwargs)
    def EvalValue(self, x):
        val = Coefficient_Evaluator.EvalValue(self, x).reshape(3,3)
        val = self.apply_factor(x, val)
        return val[1, 1]

class eps_o_r(object):
    def apply_factor(self, x, v):
        v = - v * epsilon0 * self.omega * self.omega / x[0] * self.cnorm
        if self.real:  return v.real
        else: return v.imag
class eps_x_r(object):
    def apply_factor(self, x, v):
        v = - v * epsilon0 * self.omega * self.omega * x[0]* self.cnorm
        if self.real:  return v.real
        else: return v.imag
class eps(object):
    def apply_factor(self, x, v):
        v = - v * epsilon0 * self.omega * self.omega * self.cnorm
        if self.real:  return v.real
        else: return v.imag

class sigma_o_r(object):
    def apply_factor(self, x, v):
        v = - 1j * self.omega * v / x[0] * self.cnorm
        if self.real:  return v.real
        else: return v.imag
class sigma_x_r(object):
    def apply_factor(self, x, v):
        v = - 1j * self.omega * v * x[0] * self.cnorm
        if self.real:  return v.real
        else: return v.imag
class sigma(object):
    def apply_factor(self, x, v):
        v = - 1j * self.omega * v * self.cnorm
        if self.real:  return v.real
        else: return v.imag

#class Epsilon_o_r_rz(M_RZ, eps_o_r):
#    pass
class Epsilon_o_r_phi(M_PHI, eps_o_r):
    pass
#class Sigma_o_r_rz(M_RZ, sigma_o_r):
#    pass
class Sigma_o_r_phi(M_PHI, sigma_o_r):
    pass

class Epsilon_x_r_rz(M_RZ, eps_x_r):
    pass
class Epsilon_x_r_phi(M_PHI, eps_x_r):
    pass
class Sigma_x_r_rz(M_RZ, sigma_x_r):
    pass
class Sigma_x_r_phi(M_PHI, sigma_x_r):
    pass


class Epsilon_21(M_21, eps):
    pass

class Epsilon_12(M_12, eps):
    pass

class Sigma_12(M_12, sigma):
    pass

class Sigma_21(M_21, sigma):
    pass

class InvMu_x_r(PhysCoefficient):
   '''
      r/mu0/mur
   '''
   def __init__(self, *args, **kwargs):
       self.cnorm = kwargs.pop('cnorm')
       super(InvMu_x_r, self).__init__(*args, **kwargs)

   def EvalValue(self, x):
       v = super(InvMu_x_r, self).EvalValue(x)
       v = 1/mu0/v*x[0] * self.cnorm
       if self.real:  return v.real
       else: return v.imag

class InvMu_o_r(PhysCoefficient):
   '''
      1/mu0/mur/r
   '''
   def __init__(self, *args, **kwargs):
       self.tmode = kwargs.pop('tmode', 0.0)
       self.cnorm = kwargs.pop('cnorm')
       super(InvMu_o_r, self).__init__(*args, **kwargs)

   def EvalValue(self, x):
       v = super(InvMu_o_r, self).EvalValue(x)
       v = 1./mu0/v/x[0] * self.cnorm
       if self.real:  return v.real
       else: return v.imag

class iInvMu_m_o_r(PhysCoefficient):
   '''
      1j/mu0/mur/r
   '''
   def __init__(self, *args, **kwargs):
       self.tmode = kwargs.pop('tmode', 1.0)
       self.cnorm = kwargs.pop('cnorm')
       super(iInvMu_m_o_r, self).__init__(*args, **kwargs)

   def EvalValue(self, x):
       v = super(iInvMu_m_o_r, self).EvalValue(x)
       v = 1j/mu0/v/x[0]*self.tmode * self.cnorm
       if self.real:  return v.real
       else: return v.imag

class InvMu_m2_o_r(PhysCoefficient):
   '''
      1./mu0/mur/r/m^2
   '''
   def __init__(self, *args, **kwargs):
       self.tmode = kwargs.pop('tmode', 1.0)
       self.cnorm = kwargs.pop('cnorm')
       super(InvMu_m2_o_r, self).__init__(*args, **kwargs)

   def EvalValue(self, x):
       v = super(InvMu_m2_o_r, self).EvalValue(x)
       v = 1/mu0/v/x[0]*self.tmode*self.tmode *self.cnorm
       if self.real:  return v.real
       else: return v.imag

def domain_constraints():
   return [EM2Da_Anisotropic]

class EM2Da_Anisotropic(EM2Da_Domain):
    vt  = Vtable(data)
    #nlterms = ['epsilonr']

    def has_bf_contribution(self, kfes):
        if kfes == 0: return True
        elif kfes == 1: return True
        else: return False

    def has_mixed_contribution(self):
        return True

    def get_mixedbf_loc(self):
        '''
        r, c, and flag1, flag2 of MixedBilinearForm
           flag1 : take transpose
           flag2 : take conj
        '''
        return [(0, 1, 1, 1), (1, 0, 1, 1),]#(0, 1, -1, 1)]

    def add_bf_contribution(self, engine, a, real = True, kfes=0):
        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        e, m, s, tmode = self.vt.make_value_or_expression(self)
        if not isinstance(e, str): e = str(e)
        if not isinstance(m, str): m = str(m)
        if not isinstance(s, str): s = str(s)

        if kfes == 0: ## ND element (Epoloidal)
            if real:
                dprint1("Add ND contribution(real)" + str(self._sel_index))
            else:
                dprint1("Add ND contribution(imag)" + str(self._sel_index))
            imu_x_r = InvMu_x_r(m,  self.get_root_phys().ind_vars,
                                self._local_ns, self._global_ns,
                                real = real, cnorm=cnorm)
            s_x_r = Sigma_x_r_rz(2, s,  self.get_root_phys().ind_vars,
                              self._local_ns, self._global_ns,
                              real = real, omega = omega, cnorm=cnorm)
            e_x_r = Epsilon_x_r_rz(2, e, self.get_root_phys().ind_vars,
                                self._local_ns, self._global_ns,
                                real = real, omega = omega, cnorm=cnorm)

            self.add_integrator(engine, 'mur', imu_x_r,
                                a.AddDomainIntegrator,
                                mfem.CurlCurlIntegrator)
            self.add_integrator(engine, 'epsilonr', e_x_r,
                                a.AddDomainIntegrator,
                                mfem.VectorFEMassIntegrator)
            self.add_integrator(engine, 'sigma', s_x_r,
                                a.AddDomainIntegrator,
                                mfem.VectorFEMassIntegrator)

            if tmode != 0:
                imu_o_r_2 = InvMu_m2_o_r(m,  self.get_root_phys().ind_vars,
                                      self._local_ns, self._global_ns,
                                      real = real, tmode = tmode, cnorm=cnorm)
                self.add_integrator(engine, 'mur', imu_o_r_2,
                                    a.AddDomainIntegrator,
                                    mfem.VectorFEMassIntegrator)

        elif kfes == 1: ## H1 element (Etoroidal)
            if real:
                dprint1("Add H1 contribution(real)" + str(self._sel_index))
            else:
                dprint1("Add H1 contribution(imag)" + str(self._sel_index))
            imv_o_r_1 = InvMu_o_r(m,  self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            real = real)
            e_o_r = Epsilon_o_r_phi(e, self.get_root_phys().ind_vars,
                                self._local_ns, self._global_ns,
                                real = real, omega = omega)
            s_o_r = Sigma_o_r_phi(s,  self.get_root_phys().ind_vars,
                              self._local_ns, self._global_ns,
                              real = real, omega = omega)

            self.add_integrator(engine, 'mur', imv_o_r_1,
                                a.AddDomainIntegrator,
                                mfem.DiffusionIntegrator)
            self.add_integrator(engine, 'epsilonr', e_o_r,
                                a.AddDomainIntegrator,
                                mfem.MassIntegrator)
            self.add_integrator(engine, 'sigma', s_o_r,
                                a.AddDomainIntegrator,
                                mfem.MassIntegrator)
        else:
            pass

    def add_mix_contribution(self, engine, mbf, r, c, is_trans, real = True):
        if real:
            dprint1("Add mixed contribution(real)" + "(" + str(r) + "," + str(c) +')'
                    +str(self._sel_index))
        else:
            dprint1("Add mixed contribution(imag)" + "(" + str(r) + "," + str(c) +')'
                    +str(self._sel_index))

        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        e, m, s, tmode = self.vt.make_value_or_expression(self)

        if not isinstance(e, str): e = str(e)
        if not isinstance(m, str): m = str(m)
        if not isinstance(s, str): s = str(s)

        imv_o_r_3 = iInvMu_m_o_r(m,  self.get_root_phys().ind_vars,
                              self._local_ns, self._global_ns,
                                 real = real, tmode = tmode, cnorm=cnorm)
        if r == 1 and c == 0:
            e = Epsilon_21(2, e, self.get_root_phys().ind_vars,
                                self._local_ns, self._global_ns,
                                real = real, omega = omega, cnorm=cnorm)
            s = Sigma_21(2, s,  self.get_root_phys().ind_vars,
                              self._local_ns, self._global_ns,
                              real = real, omega = omega, cnorm=cnorm)
            # (a_vec dot u_vec, v_scalar)
            itg = mfem.MixedDotProductIntegrator
            self.add_integrator(engine, 'epsilon', e,
                                mbf.AddDomainIntegrator, itg)
            self.add_integrator(engine, 'sigma', s,
                                mbf.AddDomainIntegrator, itg)
            # (-a u_vec, div v_scalar)
            itg =  mfem.MixedVectorWeakDivergenceIntegrator
            self.add_integrator(engine, 'mur', imv_o_r_3,
                                mbf.AddDomainIntegrator, itg)

        else:
            e = Epsilon_12(2, e, self.get_root_phys().ind_vars,
                             self._local_ns, self._global_ns,
                             real = real, omega = omega, cnorm=cnorm)
            s = Sigma_12(2, s,  self.get_root_phys().ind_vars,
                           self._local_ns, self._global_ns,
                           real = real, omega = omega, cnorm=cnorm)

             #itg = mfem.MixedDotProductIntegrator
            itg = mfem.MixedVectorProductIntegrator
            self.add_integrator(engine, 'epsilon', e,
                             mbf.AddDomainIntegrator, itg)
            self.add_integrator(engine, 'sigma', s,
                             mbf.AddDomainIntegrator, itg)
            # (a grad u_scalar, v_vec)

            itg =  mfem.MixedVectorGradientIntegrator
            self.add_integrator(engine, 'mur', imv_o_r_3,
                             mbf.AddDomainIntegrator, itg)

    def add_domain_variables(self, v, n, suffix, ind_vars):
        from petram.helper.variables import add_constant

        e, m, s, tmode = self.vt.make_value_or_expression(self)

        self.do_add_matrix_expr(v, suffix, ind_vars, 'epsilonr', e)
        self.do_add_scalar_expr(v, suffix, ind_vars, 'smur', m, add_diag=3)
        self.do_add_matrix_expr(v, suffix, ind_vars, 'sigma', s)

        var = ['r', 'phi', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')

        add_constant(v, 'm_mode', suffix, np.float64(tmode),
                     domains = self._sel_index,
                     gdomain = self._global_ns)
